# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

'''
This file contains code artifacts adapted from the original implementation:
https://github.com/thu-coai/ConvLab-2/blob/master/convlab2/nlg/template/multiwoz/nlg.py
'''

import json
import os
import random

from nemo.backends.pytorch.nm import NonTrainableNM
from nemo.collections.nlp.neural_types import *
from nemo.core.neural_types import *
from nemo.utils import logging
from nemo.utils.decorators import add_port_docs

__all__ = ['TemplateNLGMultiWOZ']


# supported slot
slot2word = {
    'Fee': 'fee',
    'Addr': 'address',
    'Area': 'area',
    'Stars': 'stars',
    'Internet': 'Internet',
    'Department': 'department',
    'Choice': 'choice',
    'Ref': 'reference number',
    'Food': 'food',
    'Type': 'type',
    'Price': 'price range',
    'Stay': 'stay',
    'Phone': 'phone',
    'Post': 'postcode',
    'Day': 'day',
    'Name': 'name',
    'Car': 'car type',
    'Leave': 'leave',
    'Time': 'time',
    'Arrive': 'arrive',
    'Ticket': 'ticket',
    'Depart': 'departure',
    'People': 'people',
    'Dest': 'destination',
    'Parking': 'parking',
    'Open': 'open',
    'Id': 'Id',
    # 'TrainID': 'TrainID'
}


class TemplateNLGMultiWOZ(NonTrainableNM):
    """Generate a natural language utterance conditioned on the dialog act.
    """

    def __init__(self, mode="auto_manual", name=None):
        """
        Initializes the object
        Args:
            mode (str):
                - `auto`: templates extracted from data without manual modification, may have no match;
                - `manual`: templates with manual modification, sometimes verbose;
                - `auto_manual`: use auto templates first. When fails, use manual templates.
                both template are dict, *_template[dialog_act][slot] is a list of templates.
        """
        # Call base class constructor.
        NonTrainableNM.__init__(self, name=name)

        self.mode = mode
        template_dir = os.path.dirname(os.path.abspath(__file__))

        def read_json(filename):
            with open(filename, 'r') as f:
                return json.load(f)

        self.auto_system_template = read_json(os.path.join(template_dir, 'auto_system_template_nlg.json'))
        self.manual_system_template = read_json(os.path.join(template_dir, 'manual_system_template_nlg.json'))

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        system_acts (list): list of system actions action produced by dialog policy module
        """
        return {
            'system_acts': NeuralType(
                axes=[AxisType(kind=AxisKind.Batch, is_list=True), AxisType(kind=AxisKind.Sequence, is_list=True)],
                elements_type=StringType(),
            ),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        system_uttr (str): generated system's response
        """
        return {
            'sys_uttr': NeuralType(
                axes=[AxisType(kind=AxisKind.Batch, is_list=True)], elements_type=SystemUtterance()
            ),
        }

    def forward(self, system_acts):
        """
        Generates system response 
        Args:
            system_acts (list): system actions in the format: [['Inform', 'Train', 'Day', 'wednesday'], []] [act, domain, slot, slot_value]
        Returns:
            system_uttr (str): generated system utterance 
        """
        action = {}
        for intent, domain, slot, value in system_acts:
            k = '-'.join([domain, intent])
            action.setdefault(k, [])
            action[k].append([slot, value])
        dialog_acts = action
        mode = self.mode
        try:
            if mode == 'manual':
                system_uttr = self._manual_generate(dialog_acts, self.manual_system_template)

            elif mode == 'auto':
                system_uttr = self._auto_generate(dialog_acts, self.auto_system_template)

            elif mode == 'auto_manual':
                template1 = self.auto_system_template
                template2 = self.manual_system_template
                system_uttr = self._auto_generate(dialog_acts, template1)

                if system_uttr == 'None':
                    system_uttr = self._manual_generate(dialog_acts, template2)
            else:
                raise Exception("Invalid mode! available mode: auto, manual, auto_manual")
            # truncate a system utterance with multiple questions
            system_uttr = self.truncate_sys_response(system_uttr)
            logging.info("NLG output = System reply: %s", system_uttr)
            return system_uttr
        except Exception as e:
            logging.error('Error in processing: %s', dialog_acts)
            raise e

    def truncate_sys_response(self, sys_uttr):
        """
        Truncates system response when too many questions are asked by the system
        Args:
            sys_uttr (str): generated system response
        Returns:
            (str): updated system response
        """
        start_idx = 0
        utterances_with_period = []
        utterances_with_question_mark = []
        for idx, ch in enumerate(sys_uttr):
            if ch == '?':
                utterances_with_question_mark.append((sys_uttr[start_idx : idx + 1]).strip())
                start_idx = idx + 1
            elif ch == '.':
                utterances_with_period.append((sys_uttr[start_idx : idx + 1]).strip())
                start_idx = idx + 1

        if len(utterances_with_question_mark) > 0:
            utterances_with_question_mark = utterances_with_question_mark[:1]

        return ' '.join(utterances_with_period) + ' '.join(utterances_with_question_mark)

    def _postprocess(self, sen):
        sen_strip = sen.strip()
        sen = ''.join([val.capitalize() if i == 0 else val for i, val in enumerate(sen_strip)])
        if len(sen) > 0 and sen[-1] != '?' and sen[-1] != '.':
            sen += '.'
        sen += ' '
        return sen

    def _manual_generate(self, dialog_acts, template):
        sentences = ''
        for dialog_act, slot_value_pairs in dialog_acts.items():
            intent = dialog_act.split('-')
            if 'Select' == intent[1]:
                slot2values = {}
                for slot, value in slot_value_pairs:
                    slot2values.setdefault(slot, [])
                    slot2values[slot].append(value)
                for slot, values in slot2values.items():
                    if slot == 'none':
                        continue
                    sentence = 'Do you prefer ' + values[0]
                    for i, value in enumerate(values[1:]):
                        if i == (len(values) - 2):
                            sentence += ' or ' + value
                        else:
                            sentence += ' , ' + value
                    sentence += ' {} ? '.format(slot2word[slot])
                    sentences += sentence
            elif 'Request' == intent[1]:
                for slot, value in slot_value_pairs:
                    if dialog_act not in template or slot not in template[dialog_act]:
                        sentence = 'What is the {} of {} ? '.format(slot.lower(), dialog_act.split('-')[0].lower())
                        sentences += sentence
                    else:
                        sentence = random.choice(template[dialog_act][slot])
                        sentence = self._postprocess(sentence)
                        sentences += sentence
            elif 'general' == intent[0] and dialog_act in template:
                sentence = random.choice(template[dialog_act]['none'])
                sentence = self._postprocess(sentence)
                sentences += sentence
            else:
                for slot, value in slot_value_pairs:
                    if value in ["do nt care", "do n't care", "dontcare"]:
                        sentence = 'I don\'t care about the {} of the {}'.format(
                            slot.lower(), dialog_act.split('-')[0].lower()
                        )
                    elif dialog_act in template and slot in template[dialog_act]:
                        sentence = random.choice(template[dialog_act][slot])
                        sentence = sentence.replace('#{}-{}#'.format(dialog_act.upper(), slot.upper()), str(value))
                    else:
                        if slot in slot2word:
                            sentence = 'The {} is {} . '.format(slot2word[slot], str(value))
                        else:
                            sentence = ''
                    sentence = self._postprocess(sentence)
                    sentences += sentence
        return sentences.strip()

    def _auto_generate(self, dialog_acts, template):
        sentences = ''
        for dialog_act, slot_value_pairs in dialog_acts.items():
            key = ''
            for s, v in sorted(slot_value_pairs, key=lambda x: x[0]):
                key += s + ';'
            if dialog_act in template and key in template[dialog_act]:
                sentence = random.choice(template[dialog_act][key])
                if 'Request' in dialog_act or 'general' in dialog_act:
                    sentence = self._postprocess(sentence)
                    sentences += sentence
                else:
                    for s, v in sorted(slot_value_pairs, key=lambda x: x[0]):
                        if v != 'none':
                            sentence = sentence.replace('#{}-{}#'.format(dialog_act.upper(), s.upper()), v, 1)
                    sentence = self._postprocess(sentence)
                    sentences += sentence
            else:
                return 'None'
        return sentences.strip()



# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
from torch import nn
from torch.nn import functional as F

import nemo
from nemo.backends.pytorch import nm as nemo_nm
from nemo.backends.pytorch.nm import DataLayerNM, LossNM
from nemo.collections.asr.parts import AudioDataset, WaveformFeaturizer
from nemo.collections.tts.parts import fastspeech, fastspeech_transformer
from nemo.core.neural_types import AudioSignal, EmbeddedTextType, LengthsType, MaskType, MelSpectrogramType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['FastSpeechDataLayer', 'FastSpeech', 'FastSpeechLoss']


class FastSpeechDataLayer(DataLayerNM):
    """Data Layer for Fast Speech model.

    Basically, replicated behavior from AudioToText Data Layer, zipped with ground truth durations for additional loss.

    Args:
        manifest_filepath (str): Dataset parameter.
            Path to JSON containing data.
        durs_dir (str): Path to durations arrays directory.
        labels (list): Dataset parameter.
            List of characters that can be output by the ASR model.
            For Jasper, this is the 28 character set {a-z '}. The CTC blank
            symbol is automatically added later for models using ctc.
        batch_size (int): batch size
        sample_rate (int): Target sampling rate for data. Audio files will be
            resampled to sample_rate if it is not already.
            Defaults to 16000.
        int_values (bool): Bool indicating whether the audio file is saved as
            int data or float data.
            Defaults to False.
        eos_id (id): Dataset parameter.
            End of string symbol id used for seq2seq models.
            Defaults to None.
        min_duration (float): Dataset parameter.
            All training files which have a duration less than min_duration
            are dropped. Note: Duration is read from the manifest JSON.
            Defaults to 0.1.
        max_duration (float): Dataset parameter.
            All training files which have a duration more than max_duration
            are dropped. Note: Duration is read from the manifest JSON.
            Defaults to None.
        normalize_transcripts (bool): Dataset parameter.
            Whether to use automatic text cleaning.
            It is highly recommended to manually clean text for best results.
            Defaults to True.
        trim_silence (bool): Whether to use trim silence from beginning and end
            of audio signal using librosa.effects.trim().
            Defaults to False.
        load_audio (bool): Dataset parameter.
            Controls whether the dataloader loads the audio signal and
            transcript or just the transcript.
            Defaults to True.
        drop_last (bool): See PyTorch DataLoader.
            Defaults to False.
        shuffle (bool): See PyTorch DataLoader.
            Defaults to True.
        num_workers (int): See PyTorch DataLoader.
            Defaults to 0.
        perturb_config (dict): Currently disabled.

    """

    @property
    @add_port_docs
    def output_ports(self):
        """Returns definitions of module output ports."""
        return dict(
            audio=NeuralType(('B', 'T'), AudioSignal(freq=self.sample_rate)),
            audio_len=NeuralType(tuple('B'), LengthsType()),
            text=NeuralType(('B', 'T'), EmbeddedTextType()),
            text_pos=NeuralType(('B', 'T'), MaskType()),
            dur_true=NeuralType(('B', 'T'), LengthsType()),
        )

    def __init__(
        self,
        manifest_filepath,
        durs_dir,
        labels,
        batch_size,
        sample_rate=16000,
        int_values=False,
        bos_id=None,
        eos_id=None,
        pad_id=None,
        min_duration=0.1,
        max_duration=None,
        normalize_transcripts=True,
        trim_silence=False,
        load_audio=True,
        drop_last=False,
        shuffle=True,
        num_workers=0,
    ):
        super().__init__()

        # Set up dataset.
        self._featurizer = WaveformFeaturizer(sample_rate=sample_rate, int_values=int_values, augmentor=None)
        dataset_params = {
            'manifest_filepath': manifest_filepath,
            'labels': labels,
            'featurizer': self._featurizer,
            'max_duration': max_duration,
            'min_duration': min_duration,
            'normalize': normalize_transcripts,
            'trim': trim_silence,
            'bos_id': bos_id,
            'eos_id': eos_id,
            'load_audio': load_audio,
        }
        audio_dataset = AudioDataset(**dataset_params)
        self._dataset = fastspeech.FastSpeechDataset(audio_dataset, durs_dir)
        self._pad_id = pad_id
        self.sample_rate = sample_rate

        sampler = None
        if self._placement == nemo.core.DeviceType.AllGpu:
            sampler = torch.utils.data.distributed.DistributedSampler(self._dataset)

        self._dataloader = torch.utils.data.DataLoader(
            dataset=self._dataset,
            batch_size=batch_size,
            collate_fn=self._collate,
            drop_last=drop_last,
            shuffle=shuffle if sampler is None else False,
            sampler=sampler,
            num_workers=num_workers,
        )

    def _collate(self, batch):
        def merge(tensors, value=0.0, dtype=torch.float):
            max_len = max(tensor.shape[0] for tensor in tensors)
            new_tensors = []
            for tensor in tensors:
                pad = (2 * len(tensor.shape)) * [0]
                pad[-1] = max_len - tensor.shape[0]
                new_tensors.append(F.pad(tensor, pad=pad, value=value))
            return torch.stack(new_tensors).to(dtype=dtype)

        def make_pos(lengths):
            return merge([torch.arange(length) + 1 for length in lengths], value=0, dtype=torch.int64)

        batch = {key: [example[key] for example in batch] for key in batch[0]}

        audio = merge(batch['audio'])
        audio_len = torch.tensor(batch['audio_len'])
        text = merge(batch['text'], value=self._pad_id or 0, dtype=torch.long)
        text_pos = make_pos(batch.pop('text_len'))
        dur_true = merge(batch['dur_true'])

        assert text.shape == text_pos.shape
        assert text.shape == dur_true.shape

        return audio, audio_len, text, text_pos, dur_true

    def __len__(self) -> int:
        return len(self._dataset)

    @property
    def dataset(self) -> Optional[torch.utils.data.Dataset]:
        return None

    @property
    def data_iterator(self) -> Optional[torch.utils.data.DataLoader]:
        return self._dataloader


class FastSpeech(nemo_nm.TrainableNM):
    """FastSpeech Model.

    Attributes:
        decoder_output_size: Output size for decoder.
        n_mels: Number of features for mel spectrogram.
        max_seq_len: Maximum length of input sequence.
        word_vec_dim: Dimensionality of word embedding vector.
        encoder_n_layer: Number of layers for encoder.
        encoder_head: Number of heads for encoder.
        encoder_conv1d_filter_size: Filter size for encoder convolutions.
        decoder_n_layer: Number of layers for decoder.
        decoder_head: Number of heads for decoder.
        decoder_conv1d_filter_size: Filter size for decoder convolutions.
        fft_conv1d_kernel: Kernel size for FFT.
        fft_conv1d_padding: Padding for FFT.
        encoder_output_size: Output size for encoder.
        duration_predictor_filter_size: Predictor filter size.
        duration_predictor_kernel_size: Predictor kernel size.
        dropout: Dropout probability.
        Alpha: Predictor loss coeficient.

    """

    @property
    @add_port_docs
    def input_ports(self):
        """Returns definitions of module input ports."""
        return dict(
            text=NeuralType(('B', 'T'), EmbeddedTextType()),
            text_pos=NeuralType(('B', 'T'), MaskType()),
            mel_true=NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
            dur_true=NeuralType(('B', 'T'), LengthsType()),
        )

    @property
    @add_port_docs
    def output_ports(self):
        """Returns definitions of module output ports."""
        return dict(
            mel_pred=NeuralType(('B', 'D', 'T'), MelSpectrogramType()), dur_pred=NeuralType(('B', 'T'), LengthsType()),
        )

    def __init__(
        self,
        decoder_output_size: int,
        n_mels: int,
        max_seq_len: int,
        word_vec_dim: int,
        encoder_n_layer: int,
        encoder_head: int,
        encoder_conv1d_filter_size: int,
        decoder_n_layer: int,
        decoder_head: int,
        decoder_conv1d_filter_size: int,
        fft_conv1d_kernel: int,
        fft_conv1d_padding: int,
        encoder_output_size: int,
        duration_predictor_filter_size: int,
        duration_predictor_kernel_size: int,
        dropout: float,
        alpha: float,
        n_src_vocab: int,
        pad_id: int,
    ):
        super().__init__()

        self.encoder = fastspeech_transformer.FastSpeechTransformerEncoder(
            len_max_seq=max_seq_len,
            d_word_vec=word_vec_dim,
            n_layers=encoder_n_layer,
            n_head=encoder_head,
            d_k=64,
            d_v=64,
            d_model=word_vec_dim,
            d_inner=encoder_conv1d_filter_size,
            fft_conv1d_kernel=fft_conv1d_kernel,
            fft_conv1d_padding=fft_conv1d_padding,
            dropout=dropout,
            n_src_vocab=n_src_vocab,
            pad_id=pad_id,
        ).to(self._device)
        self.length_regulator = fastspeech.LengthRegulator(
            encoder_output_size, duration_predictor_filter_size, duration_predictor_kernel_size, dropout
        ).to(self._device)

        self.decoder = fastspeech_transformer.FastSpeechTransformerDecoder(
            len_max_seq=max_seq_len,
            d_word_vec=word_vec_dim,
            n_layers=decoder_n_layer,
            n_head=decoder_head,
            d_k=64,
            d_v=64,
            d_model=word_vec_dim,
            d_inner=decoder_conv1d_filter_size,
            fft_conv1d_kernel=fft_conv1d_kernel,
            fft_conv1d_padding=fft_conv1d_padding,
            dropout=dropout,
            pad_id=pad_id,
        ).to(self._device)

        self.mel_linear = nn.Linear(decoder_output_size, n_mels, bias=True).to(self._device)
        self.alpha = alpha

    def forward(self, text, text_pos, mel_true=None, dur_true=None):
        encoder_output, encoder_mask = self.encoder(text, text_pos)

        if self.training:
            mel_max_length = mel_true.shape[2]
            length_regulator_output, decoder_pos, dur_pred = self.length_regulator(
                encoder_output, encoder_mask, dur_true, self.alpha, mel_max_length
            )

            assert length_regulator_output.shape[1] <= mel_max_length

        else:
            length_regulator_output, decoder_pos, dur_pred = self.length_regulator(
                encoder_output, encoder_mask, alpha=self.alpha
            )

        decoder_output, decoder_mask = self.decoder(length_regulator_output, decoder_pos)
        mel_pred = self.mel_linear(decoder_output).transpose(1, 2)

        assert mel_pred.shape[2] == dur_true.sum(-1).max()
        assert mel_true.shape[2] == dur_true.sum(-1).max()

        return mel_pred, dur_pred


class FastSpeechLoss(LossNM):
    """Neural Module Wrapper for Fast Speech Loss.

    Calculates final loss as sum of two: MSE for mel spectrograms and MSE for durations.

    """

    @property
    @add_port_docs
    def input_ports(self):
        """Returns definitions of module input ports."""
        return dict(
            mel_true=NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
            mel_pred=NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
            dur_true=NeuralType(('B', 'T'), LengthsType()),
            dur_pred=NeuralType(('B', 'T'), LengthsType()),
            text_pos=NeuralType(('B', 'T'), MaskType()),
        )

    @property
    @add_port_docs
    def output_ports(self):
        """Returns definitions of module output ports."""
        return dict(loss=NeuralType(None))

    def _loss_function(self, **kwargs):
        return self._loss(*(kwargs.values()))

    @staticmethod
    def _loss(
        mel_true, mel_pred, dur_true, dur_pred, text_pos,
    ):
        """Do the actual math in FastSpeech loss calculation.

        Args:
            mel_true: Ground truth mel spectrogram features (BTC, float).
            mel_pred: Predicted mel spectrogram features (BTC, float).
            dur_true: Ground truth durations (BQ, float).
            dur_pred: Predicted log-normalized durations (BQ, float).

        Returns:
            Single 0-dim loss tensor.

        """

        mel_loss = F.mse_loss(mel_pred, mel_true, reduction='none')
        mel_loss *= mel_true.ne(0).float()
        mel_loss = mel_loss.mean()

        dur_loss = F.mse_loss(dur_pred, (dur_true + 1).log(), reduction='none')
        dur_loss *= text_pos.ne(0).float()
        dur_loss = dur_loss.mean()

        loss = mel_loss + dur_loss

        return loss



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import youtokentome as yttm

from nemo.collections.nlp.data.tokenizers.tokenizer_spec import TokenizerSpec

__all__ = ['YouTokenToMeTokenizer']


class YouTokenToMeTokenizer(TokenizerSpec):
    def __init__(self, model_path):
        self.tokenizer = yttm.BPE(model=model_path)
        self.vocab_size = len(self.tokenizer.vocab())
        self.special_tokens = self.tokens_to_ids(["<PAD>", "<UNK>", "<BOS>", "<EOS>"])

    def text_to_tokens(self, text):
        return self.tokenizer.encode(text, output_type=yttm.OutputType.SUBWORD)

    def tokens_to_text(self, tokens):
        return self.ids_to_text(self.tokens_to_ids(tokens))

    def text_to_ids(self, text):
        return self.tokenizer.encode(text, output_type=yttm.OutputType.ID)

    def ids_to_text(self, ids):
        ids_ = [id_ for id_ in ids if id_ not in self.special_tokens]
        return self.tokenizer.decode([ids_])[0]

    def tokens_to_ids(self, tokens):
        return [self.tokenizer.subword_to_id(token) for token in tokens]

    def ids_to_tokens(self, ids):
        ids_ = [id_ for id_ in ids if id_ not in self.special_tokens]
        return [self.tokenizer.id_to_subword(id_) for id_ in ids_]

    @property
    def pad_id(self):
        return self.tokenizer.subword_to_id("<PAD>")

    @property
    def bos_id(self):
        return self.tokenizer.subword_to_id("<BOS>")

    @property
    def eos_id(self):
        return self.tokenizer.subword_to_id("<EOS>")



# =============================================================================
# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import pytest

from nemo import logging
from nemo.core import DeviceType, NeuralModuleFactory


def pytest_addoption(parser):
    """ Additional command-line arguments passed to pytest. For now: --cpu """
    parser.addoption('--cpu', action='store_true', help="pass that argument to use CPU during testing (default: GPU)")


@pytest.fixture
def device(request):
    """ Simple fixture returning string denoting the device [CPU | GPU] """
    if request.config.getoption("--cpu"):
        return "CPU"
    else:
        return "GPU"


@pytest.fixture(scope="class")
def neural_factory(request):
    """ Fixture creating a Neural Factory object parametrized by the command line --cpu argument """
    # Get flag.
    if request.config.getoption("--cpu"):
        device = DeviceType.CPU
    else:
        device = DeviceType.GPU
    # Initialize the default Neural Factory - on GPU.
    request.cls.nf = NeuralModuleFactory(placement=device)

    # Print standard header.
    logging.info("Using {} during testing".format(request.cls.nf.placement))


@pytest.fixture(autouse=True)
def run_only_on_device_fixture(request, device):
    if request.node.get_closest_marker('run_only_on'):
        if request.node.get_closest_marker('run_only_on').args[0] != device:
            pytest.skip('skipped on this device: {}'.format(device))


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "run_only_on(device): runs the test only on a given device [CPU | GPU]",
    )



# Taken straight from Patter https://github.com/ryanleary/patter
# TODO: review, and copyright and fix/add comments
import random

import librosa
import numpy as np
import soundfile as sf


class AudioSegment(object):
    """Monaural audio segment abstraction.
    :param samples: Audio samples [num_samples x num_channels].
    :type samples: ndarray.float32
    :param sample_rate: Audio sample rate.
    :type sample_rate: int
    :raises TypeError: If the sample data type is not float or int.
    """

    def __init__(self, samples, sample_rate, target_sr=None, trim=False, trim_db=60):
        """Create audio segment from samples.
        Samples are convert float32 internally, with int scaled to [-1, 1].
        """
        samples = self._convert_samples_to_float32(samples)
        if target_sr is not None and target_sr != sample_rate:
            samples = librosa.core.resample(samples, sample_rate, target_sr)
            sample_rate = target_sr
        if trim:
            samples, _ = librosa.effects.trim(samples, trim_db)
        self._samples = samples
        self._sample_rate = sample_rate
        if self._samples.ndim >= 2:
            self._samples = np.mean(self._samples, 1)

    def __eq__(self, other):
        """Return whether two objects are equal."""
        if type(other) is not type(self):
            return False
        if self._sample_rate != other._sample_rate:
            return False
        if self._samples.shape != other._samples.shape:
            return False
        if np.any(self.samples != other._samples):
            return False
        return True

    def __ne__(self, other):
        """Return whether two objects are unequal."""
        return not self.__eq__(other)

    def __str__(self):
        """Return human-readable representation of segment."""
        return "%s: num_samples=%d, sample_rate=%d, duration=%.2fsec, rms=%.2fdB" % (
            type(self),
            self.num_samples,
            self.sample_rate,
            self.duration,
            self.rms_db,
        )

    @staticmethod
    def _convert_samples_to_float32(samples):
        """Convert sample type to float32.
        Audio sample type is usually integer or float-point.
        Integers will be scaled to [-1, 1] in float32.
        """
        float32_samples = samples.astype('float32')
        if samples.dtype in np.sctypes['int']:
            bits = np.iinfo(samples.dtype).bits
            float32_samples *= 1.0 / 2 ** (bits - 1)
        elif samples.dtype in np.sctypes['float']:
            pass
        else:
            raise TypeError("Unsupported sample type: %s." % samples.dtype)
        return float32_samples

    @classmethod
    def from_file(
        cls, audio_file, target_sr=None, int_values=False, offset=0, duration=0, trim=False,
    ):
        """
        Load a file supported by librosa and return as an AudioSegment.
        :param audio_file: path of file to load
        :param target_sr: the desired sample rate
        :param int_values: if true, load samples as 32-bit integers
        :param offset: offset in seconds when loading audio
        :param duration: duration in seconds when loading audio
        :return: numpy array of samples
        """
        with sf.SoundFile(audio_file, 'r') as f:
            dtype = 'int32' if int_values else 'float32'
            sample_rate = f.samplerate
            if offset > 0:
                f.seek(int(offset * sample_rate))
            if duration > 0:
                samples = f.read(int(duration * sample_rate), dtype=dtype)
            else:
                samples = f.read(dtype=dtype)

        samples = samples.transpose()
        return cls(samples, sample_rate, target_sr=target_sr, trim=trim)

    @classmethod
    def segment_from_file(cls, audio_file, target_sr=None, n_segments=0, trim=False):
        """Grabs n_segments number of samples from audio_file randomly from the
        file as opposed to at a specified offset.

        Note that audio_file can be either the file path, or a file-like object.
        """
        with sf.SoundFile(audio_file, 'r') as f:
            sample_rate = f.samplerate
            if n_segments > 0 and len(f) > n_segments:
                max_audio_start = len(f) - n_segments
                audio_start = random.randint(0, max_audio_start)
                f.seek(audio_start)
                samples = f.read(n_segments, dtype='float32')
            else:
                samples = f.read(dtype='float32')

        samples = samples.transpose()
        return cls(samples, sample_rate, target_sr=target_sr, trim=trim)

    @property
    def samples(self):
        return self._samples.copy()

    @property
    def sample_rate(self):
        return self._sample_rate

    @property
    def num_samples(self):
        return self._samples.shape[0]

    @property
    def duration(self):
        return self._samples.shape[0] / float(self._sample_rate)

    @property
    def rms_db(self):
        mean_square = np.mean(self._samples ** 2)
        return 10 * np.log10(mean_square)

    def gain_db(self, gain):
        self._samples *= 10.0 ** (gain / 20.0)

    def pad(self, pad_size, symmetric=False):
        """Add zero padding to the sample. The pad size is given in number
        of samples.
        If symmetric=True, `pad_size` will be added to both sides. If false,
        `pad_size`
        zeros will be added only to the end.
        """
        self._samples = np.pad(self._samples, (pad_size if symmetric else 0, pad_size), mode='constant',)

    def subsegment(self, start_time=None, end_time=None):
        """Cut the AudioSegment between given boundaries.
        Note that this is an in-place transformation.
        :param start_time: Beginning of subsegment in seconds.
        :type start_time: float
        :param end_time: End of subsegment in seconds.
        :type end_time: float
        :raise ValueError: If start_time or end_time is incorrectly set,
        e.g. out
                           of bounds in time.
        """
        start_time = 0.0 if start_time is None else start_time
        end_time = self.duration if end_time is None else end_time
        if start_time < 0.0:
            start_time = self.duration + start_time
        if end_time < 0.0:
            end_time = self.duration + end_time
        if start_time < 0.0:
            raise ValueError("The slice start position (%f s) is out of bounds." % start_time)
        if end_time < 0.0:
            raise ValueError("The slice end position (%f s) is out of bounds." % end_time)
        if start_time > end_time:
            raise ValueError(
                "The slice start position (%f s) is later than the end position (%f s)." % (start_time, end_time)
            )
        if end_time > self.duration:
            raise ValueError("The slice end position (%f s) is out of bounds (> %f s)" % (end_time, self.duration))
        start_sample = int(round(start_time * self._sample_rate))
        end_sample = int(round(end_time * self._sample_rate))
        self._samples = self._samples[start_sample:end_sample]



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2019, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['NeuralTypeComparisonResult']

from enum import Enum


class NeuralTypeComparisonResult(Enum):
    """The result of comparing two neural type objects for compatibility.
    When comparing A.compare_to(B):"""

    SAME = 0
    LESS = 1  # A is B
    GREATER = 2  # B is A
    DIM_INCOMPATIBLE = 3  # Resize connector might fix incompatibility
    TRANSPOSE_SAME = 4  # A transpose and/or converting between lists and tensors will make them same
    CONTAINER_SIZE_MISMATCH = 5  # A and B contain different number of elements
    INCOMPATIBLE = 6  # A and B are incompatible
    SAME_TYPE_INCOMPATIBLE_PARAMS = 7  # A and B are of the same type but parametrized differently



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

'''
This file contains code artifacts adapted from the original implementation:
https://github.com/thu-coai/ConvLab-2/blob/master/convlab2/policy/rule/multiwoz/rule_based_multiwoz_bot.py

The code is based on:
Schatzmann, Jost, et al. "Agenda-based user simulation for bootstrapping a POMDP dialogue system."
Human Language Technologies 2007: The Conference of the North American Chapter of the Association for
Computational Linguistics;
Companion Volume, Short Papers. Association for Computational Linguistics, 2007.
'''

import random
from copy import deepcopy

from nemo import logging
from nemo.backends.pytorch.nm import NonTrainableNM
from nemo.collections.nlp.data.datasets.multiwoz_dataset.dbquery import Database
from nemo.collections.nlp.data.datasets.multiwoz_dataset.multiwoz_slot_trans import REF_SYS_DA, REF_USR_DA
from nemo.collections.nlp.neural_types import *
from nemo.core.neural_types import *
from nemo.utils.decorators import add_port_docs

__all__ = ['RuleBasedDPMMultiWOZ']

SELECTABLE_SLOTS = {
    'Attraction': ['area', 'entrance fee', 'name', 'type'],
    'Hospital': ['department'],
    'Hotel': ['area', 'internet', 'name', 'parking', 'pricerange', 'stars', 'type'],
    'Restaurant': ['area', 'name', 'food', 'pricerange'],
    'Taxi': [],
    'Train': [],
    'Police': [],
}

INFORMABLE_SLOTS = [
    "Fee",
    "Addr",
    "Area",
    "Stars",
    "Internet",
    "Department",
    "Choice",
    "Ref",
    "Food",
    "Type",
    "Price",
    "Stay",
    "Phone",
    "Post",
    "Day",
    "Name",
    "Car",
    "Leave",
    "Time",
    "Arrive",
    "Ticket",
    None,
    "Depart",
    "People",
    "Dest",
    "Parking",
    "Open",
    "Id",
]

REQUESTABLE_SLOTS = ['Food', 'Area', 'Fee', 'Price', 'Type', 'Department', 'Internet', 'Parking', 'Stars', 'Type']

# Information required to finish booking, according to different domain.
booking_info = {'Train': ['People'], 'Restaurant': ['Time', 'Day', 'People'], 'Hotel': ['Stay', 'Day', 'People']}


# Judge if user has confirmed a unique choice, according to different domain
token = {'Attraction': ['Name', 'Addr', ''], 'Hotel': ['Name',]}


class RuleBasedDPMMultiWOZ(NonTrainableNM):
    """
     Rule-based bot. Implemented for Multiwoz dataset.
    """

    def __init__(self, data_dir: str, name: str = None):
        """
        Initializes the object
        Args:
            data_dir (str): path to data directory
            name: name of the modules (DEFAULT: none)
        """
        # Call base class constructor.
        NonTrainableNM.__init__(self, name=name)
        # Set init values of attributes.
        self.last_state = {}
        self.db = Database(data_dir)
        self.last_request_state = {}
        self.last_belief_state = {}
        self.recommend_flag = -1
        self.choice = ""

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            'belief_state': NeuralType(
                axes=[
                    AxisType(kind=AxisKind.Batch, is_list=True),
                    AxisType(
                        kind=DialogAxisKind.Domain, is_list=True
                    ),  # always 7 domains - but cannot set size with is_list!
                ],
                elements_type=MultiWOZBeliefState(),
            ),
            'request_state': NeuralType(
                axes=[AxisType(kind=AxisKind.Batch, is_list=True), AxisType(kind=AxisKind.Sequence, is_list=True)],
                elements_type=StringType(),
            ),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        system_acts (list): system actions in the format: [['Inform', 'Train', 'Day', 'wednesday'], []] [act, domain, slot, slot_value]
        belief_state (dict): dialogue state with slot-slot_values pairs for all domains
        """
        return {
            'belief_state': NeuralType(
                axes=[
                    AxisType(kind=AxisKind.Batch, is_list=True),
                    AxisType(
                        kind=DialogAxisKind.Domain, is_list=True
                    ),  # always 7 domains - but cannot set size with is_list!
                ],
                elements_type=MultiWOZBeliefState(),
            ),
            'system_acts': NeuralType(
                axes=[AxisType(kind=AxisKind.Batch, is_list=True), AxisType(kind=AxisKind.Sequence, is_list=True)],
                elements_type=StringType(),
            ),
        }

    def forward(self, belief_state, request_state):
        """
        Generated System Act and add it to the belief state
        Args:
            belief_state (dict): dialogue state with slot-slot_values pairs for all domains
            request_state (dict): requested slots dict
        Returns:
            belief_state (dict): updated belief state
            system_acts (list): DA(Dialog Act), in the form of {act_type1: [[slot_name_1, value_1], [slot_name_2, value_2], ...], ...}
        """

        if self.recommend_flag != -1:
            self.recommend_flag += 1

        self.kb_result = {}
        DA = {}
        user_action = self.check_diff(belief_state, request_state)

        self.last_request_state = deepcopy(request_state)
        self.last_belief_state = deepcopy(belief_state)

        for user_act in user_action:
            domain, _ = user_act.split('-')

            # Respond to general greetings
            if domain == 'general':
                self._update_greeting(user_act, DA)

            # Book taxi for user
            elif domain == 'Taxi':
                self._book_taxi(belief_state, DA)

            elif domain == 'Booking':
                pass

            # User's talking about other domain
            elif domain != "Train":
                self._update_DA(user_act, user_action, belief_state, DA)

            # Info about train
            else:
                self._update_train(user_act, user_action, belief_state, DA)

            # Judge if user want to book
            self._judge_booking(user_act, user_action, DA)

            if 'Booking-Book' in DA:
                if random.random() < 0.5:
                    DA['general-reqmore'] = []
                user_acts = []
                for user_act_ in DA:
                    if user_act_ != 'Booking-Book':
                        user_acts.append(user_act_)
                for user_act_ in user_acts:
                    del DA[user_act_]

        if DA == {}:
            DA = {'general-greet': [['none', 'none']]}
        system_acts = []
        for domain_intent, svs in DA.items():
            domain, intent = domain_intent.split('-')
            if not svs and domain == 'general':
                system_acts.append([intent, domain, 'none', 'none'])
            else:
                for slot, value in svs:
                    system_acts.append([intent, domain, slot, value])

        logging.debug("DPM output: %s", system_acts)
        logging.debug("Belief State after DPM: %s", belief_state)
        logging.debug("Request State after DPM: %s", request_state)
        return belief_state, system_acts

    def _update_greeting(self, user_act, DA):
        """ General request / inform. """
        _, intent_type = user_act.split('-')

        # Respond to goodbye
        if intent_type == 'bye':
            if 'general-bye' not in DA:
                DA['general-bye'] = []
            if random.random() < 0.3:
                if 'general-welcome' not in DA:
                    DA['general-welcome'] = []
        elif intent_type == 'thank':
            DA['general-welcome'] = []

    def _book_taxi(self, belief_state, DA):
        """ Book a taxi for user. """

        blank_info = []
        for info in ['departure', 'destination']:
            if belief_state['taxi']['semi'] == "":
                info = REF_USR_DA['Taxi'].get(info, info)
                blank_info.append(info)
        if belief_state['taxi']['semi']['leaveAt'] == "" and belief_state['taxi']['semi']['arriveBy'] == "":
            blank_info += ['Leave', 'Arrive']

        # Finish booking, tell user car type and phone number
        if len(blank_info) == 0:
            if 'Taxi-Inform' not in DA:
                DA['Taxi-Inform'] = []
            car = self.generate_car()
            phone_num = self.generate_phone_num(11)
            DA['Taxi-Inform'].append(['Car', car])
            DA['Taxi-Inform'].append(['Phone', phone_num])
            return

        # Need essential info to finish booking
        request_num = random.randint(0, 999999) % len(blank_info) + 1
        if 'Taxi-Request' not in DA:
            DA['Taxi-Request'] = []
        for i in range(request_num):
            slot = REF_USR_DA.get(blank_info[i], blank_info[i])
            DA['Taxi-Request'].append([slot, '?'])

    def _update_DA(self, user_act, user_action, belief_state, DA):
        """ Answer user's utterance about any domain other than taxi or train. """

        domain, intent_type = user_act.split('-')
        if domain.lower() not in belief_state.keys():
            return
        constraints = []
        for slot in belief_state[domain.lower()]['semi']:
            if belief_state[domain.lower()]['semi'][slot] != "":
                constraints.append([slot, belief_state[domain.lower()]['semi'][slot]])

        kb_result = self.db.query(domain.lower(), constraints)
        self.kb_result[domain] = deepcopy(kb_result)

        # Respond to user's request
        if intent_type == 'Request':
            if self.recommend_flag > 1:
                self.recommend_flag = -1
                self.choice = ""
            elif self.recommend_flag == 1:
                self.recommend_flag == 0
            if (domain + "-Inform") not in DA:
                DA[domain + "-Inform"] = []
            for slot in user_action[user_act]:
                if len(kb_result) > 0:
                    kb_slot_name = REF_SYS_DA[domain].get(slot[0], slot[0])
                    if kb_slot_name in kb_result[0]:
                        DA[domain + "-Inform"].append([slot[0], kb_result[0][kb_slot_name]])
                    else:
                        DA[domain + "-Inform"].append([slot[0], "unknown"])

        else:
            # There's no result matching user's constraint
            if len(kb_result) == 0:
                if (domain + "-NoOffer") not in DA:
                    DA[domain + "-NoOffer"] = []

                for slot in belief_state[domain.lower()]['semi']:
                    if belief_state[domain.lower()]['semi'][slot] != "" and belief_state[domain.lower()]['semi'][
                        slot
                    ] not in ["do nt care", "do n't care", "dontcare"]:
                        slot_name = REF_USR_DA[domain].get(slot, slot)
                        DA[domain + "-NoOffer"].append([slot_name, belief_state[domain.lower()]['semi'][slot]])

                p = random.random()

                # Ask user if he wants to change constraint
                if p < 0.3:
                    req_num = min(random.randint(0, 999999) % len(DA[domain + "-NoOffer"]) + 1, 3)
                    if domain + "-Request" not in DA:
                        DA[domain + "-Request"] = []
                    for i in range(req_num):
                        slot_name = REF_USR_DA[domain].get(
                            DA[domain + "-NoOffer"][i][0], DA[domain + "-NoOffer"][i][0]
                        )
                        DA[domain + "-Request"].append([slot_name, "?"])

            # There's exactly one result matching user's constraint
            # elif len(state['kb_results_dict']) == 1:
            elif len(kb_result) == 1:

                # Inform user about this result
                if (domain + "-Inform") not in DA:
                    DA[domain + "-Inform"] = []
                props = []
                for prop in belief_state[domain.lower()]['semi']:
                    props.append(prop)
                property_num = len(props)
                if property_num > 0:
                    info_num = random.randint(0, 999999) % property_num + 1
                    random.shuffle(props)
                    for i in range(info_num):
                        slot_name = REF_USR_DA[domain].get(props[i], props[i])
                        # DA[domain + "-Inform"].append([slot_name, state['kb_results_dict'][0][props[i]]])
                        DA[domain + "-Inform"].append([slot_name, kb_result[0][props[i]]])

            # There are multiple resultes matching user's constraint
            else:
                p = random.random()

                # Recommend a choice from kb_list
                if p < 0.3:
                    if (domain + "-Inform") not in DA:
                        DA[domain + "-Inform"] = []
                    if (domain + "-Recommend") not in DA:
                        DA[domain + "-Recommend"] = []
                    DA[domain + "-Inform"].append(["Choice", str(len(kb_result))])
                    idx = random.randint(0, 999999) % len(kb_result)
                    # idx = 0
                    choice = kb_result[idx]
                    if domain in ["Hotel", "Attraction", "Police", "Restaurant"]:
                        DA[domain + "-Recommend"].append(['Name', choice['name']])
                    self.recommend_flag = 0
                    self.candidate = choice
                    props = []
                    for prop in choice:
                        props.append([prop, choice[prop]])
                    prop_num = min(random.randint(0, 999999) % 3, len(props))
                    # prop_num = min(2, len(props))
                    random.shuffle(props)
                    for i in range(prop_num):
                        slot = props[i][0]
                        string = REF_USR_DA[domain].get(slot, slot)
                        if string in INFORMABLE_SLOTS:
                            DA[domain + "-Recommend"].append([string, str(props[i][1])])

                # Ask user to choose a candidate.
                elif p < 0.5:
                    prop_values = []
                    props = []
                    # for prop in state['kb_results_dict'][0]:
                    for prop in kb_result[0]:
                        # for candidate in state['kb_results_dict']:
                        for candidate in kb_result:
                            if prop not in candidate:
                                continue
                            if candidate[prop] not in prop_values:
                                prop_values.append(candidate[prop])
                        if len(prop_values) > 1:
                            props.append([prop, prop_values])
                        prop_values = []
                    random.shuffle(props)
                    idx = 0
                    while idx < len(props):
                        if props[idx][0] not in SELECTABLE_SLOTS[domain]:
                            props.pop(idx)
                            idx -= 1
                        idx += 1
                    if domain + "-Select" not in DA:
                        DA[domain + "-Select"] = []
                    for i in range(min(len(props[0][1]), 5)):
                        prop_value = REF_USR_DA[domain].get(props[0][0], props[0][0])
                        DA[domain + "-Select"].append([prop_value, props[0][1][i]])

                # Ask user for more constraint
                else:
                    reqs = []
                    for prop in belief_state[domain.lower()]['semi']:
                        if belief_state[domain.lower()]['semi'][prop] == "":
                            prop_value = REF_USR_DA[domain].get(prop, prop)
                            reqs.append([prop_value, "?"])
                    i = 0
                    while i < len(reqs):
                        if reqs[i][0] not in REQUESTABLE_SLOTS:
                            reqs.pop(i)
                            i -= 1
                        i += 1
                    random.shuffle(reqs)
                    if len(reqs) == 0:
                        return
                    req_num = min(random.randint(0, 999999) % len(reqs) + 1, 2)
                    if (domain + "-Request") not in DA:
                        DA[domain + "-Request"] = []
                    for i in range(req_num):
                        req = reqs[i]
                        req[0] = REF_USR_DA[domain].get(req[0], req[0])
                        DA[domain + "-Request"].append(req)

    def _update_train(self, user_act, user_action, belief_state, DA):
        constraints = []
        for time in ['leaveAt', 'arriveBy']:
            if belief_state['train']['semi'][time] != "":
                constraints.append([time, belief_state['train']['semi'][time]])

        if len(constraints) == 0:
            p = random.random()
            if 'Train-Request' not in DA:
                DA['Train-Request'] = []
            if p < 0.33:
                DA['Train-Request'].append(['Leave', '?'])
            elif p < 0.66:
                DA['Train-Request'].append(['Arrive', '?'])
            else:
                DA['Train-Request'].append(['Leave', '?'])
                DA['Train-Request'].append(['Arrive', '?'])

        if 'Train-Request' not in DA:
            DA['Train-Request'] = []
        for prop in ['day', 'destination', 'departure']:
            if belief_state['train']['semi'][prop] == "":
                slot = REF_USR_DA['Train'].get(prop, prop)
                DA["Train-Request"].append([slot, '?'])
            else:
                constraints.append([prop, belief_state['train']['semi'][prop]])

        kb_result = self.db.query('train', constraints)
        self.kb_result['Train'] = deepcopy(kb_result)

        if user_act == 'Train-Request':
            del DA['Train-Request']
            if 'Train-Inform' not in DA:
                DA['Train-Inform'] = []
            for slot in user_action[user_act]:
                # Train_DA_MAP = {'Duration': "Time", 'Price': 'Ticket', 'TrainID': 'Id'}
                # slot[0] = Train_DA_MAP.get(slot[0], slot[0])
                slot_name = REF_SYS_DA['Train'].get(slot[0], slot[0])
                try:
                    DA['Train-Inform'].append([slot[0], kb_result[0][slot_name]])
                except Exception:
                    pass
            return
        if len(kb_result) == 0:
            if 'Train-NoOffer' not in DA:
                DA['Train-NoOffer'] = []
            for prop in constraints:
                DA['Train-NoOffer'].append([REF_USR_DA['Train'].get(prop[0], prop[0]), prop[1]])
            if 'Train-Request' in DA:
                del DA['Train-Request']
        elif len(kb_result) >= 1:
            if len(constraints) < 4:
                return
            if 'Train-Request' in DA:
                del DA['Train-Request']
            if 'Train-OfferBook' not in DA:
                DA['Train-OfferBook'] = []
            for prop in constraints:
                DA['Train-OfferBook'].append([REF_USR_DA['Train'].get(prop[0], prop[0]), prop[1]])

    def _judge_booking(self, user_act, user_action, DA):
        """ If user want to book, return a ref number. """
        if self.recommend_flag > 1:
            self.recommend_flag = -1
            self.choice = ""
        elif self.recommend_flag == 1:
            self.recommend_flag == 0
        domain, _ = user_act.split('-')
        for slot in user_action[user_act]:
            if domain in booking_info and slot[0] in booking_info[domain]:
                if 'Booking-Book' not in DA:
                    if domain in self.kb_result and len(self.kb_result[domain]) > 0:
                        if 'Ref' in self.kb_result[domain][0]:
                            DA['Booking-Book'] = [["Ref", self.kb_result[domain][0]['Ref']]]
                        else:
                            DA['Booking-Book'] = [["Ref", "N/A"]]

    def check_diff(self, belief_state, request_state):
        user_action = {}
        if self.last_belief_state == {} and self.last_request_state == {}:
            for domain in belief_state:
                for slot in belief_state[domain]['book']:
                    if slot != 'booked' and belief_state[domain]['book'][slot] != '':
                        if (domain.capitalize() + "-Inform") not in user_action:
                            user_action[domain.capitalize() + "-Inform"] = []
                        if [
                            REF_USR_DA[domain.capitalize()].get(slot, slot),
                            belief_state[domain]['book'][slot],
                        ] not in user_action[domain.capitalize() + "-Inform"]:
                            user_action[domain.capitalize() + "-Inform"].append(
                                [REF_USR_DA[domain.capitalize()].get(slot, slot), belief_state[domain]['book'][slot],]
                            )
                for slot in belief_state[domain]['semi']:
                    if belief_state[domain]['semi'][slot] != "":
                        if (domain.capitalize() + "-Inform") not in user_action:
                            user_action[domain.capitalize() + "-Inform"] = []
                        if [
                            REF_USR_DA[domain.capitalize()].get(slot, slot),
                            belief_state[domain]['semi'][slot],
                        ] not in user_action[domain.capitalize() + "-Inform"]:
                            user_action[domain.capitalize() + "-Inform"].append(
                                [REF_USR_DA[domain.capitalize()].get(slot, slot), belief_state[domain]['semi'][slot],]
                            )
            for domain in request_state:
                for slot in request_state[domain]:
                    if (domain.capitalize() + "-Request") not in user_action:
                        user_action[domain.capitalize() + "-Request"] = []
                    if [REF_USR_DA[domain].get(slot, slot), '?'] not in user_action[domain.capitalize() + "-Request"]:
                        user_action[domain.capitalize() + "-Request"].append([REF_USR_DA[domain].get(slot, slot), '?'])

        else:
            for domain in belief_state:
                for slot in belief_state[domain]['book']:
                    if (
                        slot != 'booked'
                        and belief_state[domain]['book'][slot] != self.last_belief_state[domain]['book'][slot]
                    ):
                        if (domain.capitalize() + "-Inform") not in user_action:
                            user_action[domain.capitalize() + "-Inform"] = []
                        if [
                            REF_USR_DA[domain.capitalize()].get(slot, slot),
                            belief_state[domain]['book'][slot],
                        ] not in user_action[domain.capitalize() + "-Inform"]:
                            user_action[domain.capitalize() + "-Inform"].append(
                                [REF_USR_DA[domain.capitalize()].get(slot, slot), belief_state[domain]['book'][slot],]
                            )
                for slot in belief_state[domain]['semi']:
                    if (
                        belief_state[domain]['semi'][slot] != self.last_belief_state[domain]['semi'][slot]
                        and belief_state[domain]['semi'][slot] != ''
                    ):
                        if (domain.capitalize() + "-Inform") not in user_action:
                            user_action[domain.capitalize() + "-Inform"] = []
                        if [
                            REF_USR_DA[domain.capitalize()].get(slot, slot),
                            belief_state[domain]['semi'][slot],
                        ] not in user_action[domain.capitalize() + "-Inform"]:
                            user_action[domain.capitalize() + "-Inform"].append(
                                [REF_USR_DA[domain.capitalize()].get(slot, slot), belief_state[domain]['semi'][slot],]
                            )
            for domain in request_state:
                for slot in request_state[domain]:
                    if (domain not in self.last_request_state) or (slot not in self.last_request_state[domain]):
                        if (domain.capitalize() + "-Request") not in user_action:
                            user_action[domain.capitalize() + "-Request"] = []
                        if [REF_USR_DA[domain.capitalize()].get(slot, slot), '?'] not in user_action[
                            domain.capitalize() + "-Request"
                        ]:
                            user_action[domain.capitalize() + "-Request"].append(
                                [REF_USR_DA[domain.capitalize()].get(slot, slot), '?']
                            )
        return user_action

    def generate_phone_num(self, length):
        """ Generate phone number."""
        string = ""
        while len(string) < length:
            string += '0123456789'[random.randint(0, 999999) % 10]
        return string

    def generate_car(self):
        """ Generate a car for taxi booking. """
        car_types = ["toyota", "skoda", "bmw", "honda", "ford", "audi", "lexus", "volvo", "volkswagen", "tesla"]
        p = random.randint(0, 999999) % len(car_types)
        return car_types[p]



# Copyright (C) NVIDIA CORPORATION. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Optional, Tuple

import torch
import torch.nn as nn
from torch import Tensor

jasper_activations = {
    "hardtanh": nn.Hardtanh,
    "relu": nn.ReLU,
    "selu": nn.SELU,
}


def init_weights(m, mode='xavier_uniform'):
    if isinstance(m, MaskedConv1d):
        init_weights(m.conv, mode)
    if isinstance(m, (nn.Conv1d, nn.Linear)):
        if mode == 'xavier_uniform':
            nn.init.xavier_uniform_(m.weight, gain=1.0)
        elif mode == 'xavier_normal':
            nn.init.xavier_normal_(m.weight, gain=1.0)
        elif mode == 'kaiming_uniform':
            nn.init.kaiming_uniform_(m.weight, nonlinearity="relu")
        elif mode == 'kaiming_normal':
            nn.init.kaiming_normal_(m.weight, nonlinearity="relu")
        else:
            raise ValueError("Unknown Initialization mode: {0}".format(mode))
    elif isinstance(m, nn.BatchNorm1d):
        if m.track_running_stats:
            m.running_mean.zero_()
            m.running_var.fill_(1)
            m.num_batches_tracked.zero_()
        if m.affine:
            nn.init.ones_(m.weight)
            nn.init.zeros_(m.bias)


def compute_new_kernel_size(kernel_size, kernel_width):
    new_kernel_size = max(int(kernel_size * kernel_width), 1)
    # If kernel is even shape, round up to make it odd
    if new_kernel_size % 2 == 0:
        new_kernel_size += 1
    return new_kernel_size


def get_same_padding(kernel_size, stride, dilation):
    if stride > 1 and dilation > 1:
        raise ValueError("Only stride OR dilation may be greater than 1")
    if dilation > 1:
        return (dilation * kernel_size) // 2 - 1
    return kernel_size // 2


class StatsPoolLayer(nn.Module):
    def __init__(self, gram=False, super_vector=False):
        super().__init__()
        self.gram = gram
        self.super = super_vector

    def forward(self, encoder_output):

        mean = encoder_output.mean(dim=-1)  # Time Axis
        std = encoder_output.std(dim=-1)

        pooled = torch.cat([mean, std], dim=-1)

        if self.gram:
            time_len = encoder_output.shape[-1]
            # encoder_output = encoder_output
            cov = encoder_output.bmm(encoder_output.transpose(2, 1))  # cov matrix
            cov = cov.view(cov.shape[0], -1) / (time_len)

        if self.gram and not self.super:
            return cov

        if self.super and self.gram:
            pooled = torch.cat([pooled, cov], dim=-1)

        return pooled


class MaskedConv1d(nn.Module):
    __constants__ = ["use_conv_mask", "real_out_channels", "heads"]

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        heads=-1,
        bias=False,
        use_mask=True,
    ):
        super(MaskedConv1d, self).__init__()

        if not (heads == -1 or groups == in_channels):
            raise ValueError("Only use heads for depthwise convolutions")

        self.real_out_channels = out_channels
        if heads != -1:
            in_channels = heads
            out_channels = heads
            groups = heads

        self.conv = nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
        )
        self.use_mask = use_mask
        self.heads = heads

    def get_seq_len(self, lens):
        return (
            lens + 2 * self.conv.padding[0] - self.conv.dilation[0] * (self.conv.kernel_size[0] - 1) - 1
        ) / self.conv.stride[0] + 1

    def forward(self, x, lens):
        if self.use_mask:
            lens = lens.to(dtype=torch.long)
            max_len = x.size(2)
            mask = torch.arange(max_len).to(lens.device).expand(len(lens), max_len) >= lens.unsqueeze(1)
            x = x.masked_fill(mask.unsqueeze(1).to(device=x.device), 0)
            # del mask
            lens = self.get_seq_len(lens)

        sh = x.shape
        if self.heads != -1:
            x = x.view(-1, self.heads, sh[-1])

        out = self.conv(x)

        if self.heads != -1:
            out = out.view(sh[0], self.real_out_channels, -1)

        return out, lens


class GroupShuffle(nn.Module):
    def __init__(self, groups, channels):
        super(GroupShuffle, self).__init__()

        self.groups = groups
        self.channels_per_group = channels // groups

    def forward(self, x):
        sh = x.shape

        x = x.view(-1, self.groups, self.channels_per_group, sh[-1])

        x = torch.transpose(x, 1, 2).contiguous()

        x = x.view(-1, self.groups * self.channels_per_group, sh[-1])

        return x


class SqueezeExcite(nn.Module):
    def __init__(
        self,
        channels: int,
        reduction_ratio: int,
        context_window: int = -1,
        interpolation_mode: str = 'nearest',
        activation: Optional[Callable] = None,
    ):
        """
        Squeeze-and-Excitation sub-module.

        Args:
            channels: Input number of channels.
            reduction_ratio: Reduction ratio for "squeeze" layer.
            context_window: Integer number of timesteps that the context
                should be computed over, using stride 1 average pooling.
                If value < 1, then global context is computed.
            interpolation_mode: Interpolation mode of timestep dimension.
                Used only if context window is > 1.
                The modes available for resizing are: `nearest`, `linear` (3D-only),
                `bilinear`, `area`
            activation: Intermediate activation function used. Must be a
                callable activation function.
        """
        super(SqueezeExcite, self).__init__()
        self.context_window = int(context_window)
        self.interpolation_mode = interpolation_mode

        if self.context_window <= 0:
            self.pool = nn.AdaptiveAvgPool1d(1)  # context window = T
        else:
            self.pool = nn.AvgPool1d(self.context_window, stride=1)

        if activation is None:
            activation = nn.ReLU(inplace=True)

        self.fc = nn.Sequential(
            nn.Linear(channels, channels // reduction_ratio, bias=False),
            activation,
            nn.Linear(channels // reduction_ratio, channels, bias=False),
        )

    def forward(self, x):
        batch, channels, timesteps = x.size()
        y = self.pool(x)  # [B, C, T - context_window + 1]
        y = y.transpose(1, 2)  # [B, T - context_window + 1, C]
        y = self.fc(y)  # [B, T - context_window + 1, C]
        y = y.transpose(1, 2)  # [B, C, T - context_window + 1]

        if self.context_window > 0:
            y = torch.nn.functional.interpolate(y, size=timesteps, mode=self.interpolation_mode)

        y = torch.sigmoid(y)

        return x * y


class Swish(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)


class JasperBlock(nn.Module):
    __constants__ = ["conv_mask", "separable", "residual_mode", "res", "mconv"]

    def __init__(
        self,
        inplanes,
        planes,
        repeat=3,
        kernel_size=11,
        kernel_size_factor=1,
        stride=1,
        dilation=1,
        padding='same',
        dropout=0.2,
        activation=None,
        residual=True,
        groups=1,
        separable=False,
        heads=-1,
        normalization="batch",
        norm_groups=1,
        residual_mode='add',
        residual_panes=[],
        conv_mask=False,
        se=False,
        se_reduction_ratio=16,
        se_context_window=None,
        se_interpolation_mode='nearest',
        stride_last=False,
    ):
        super(JasperBlock, self).__init__()

        if padding != "same":
            raise ValueError("currently only 'same' padding is supported")

        kernel_size_factor = float(kernel_size_factor)
        if type(kernel_size) in (list, tuple):
            kernel_size = [compute_new_kernel_size(k, kernel_size_factor) for k in kernel_size]
        else:
            kernel_size = compute_new_kernel_size(kernel_size, kernel_size_factor)

        padding_val = get_same_padding(kernel_size[0], stride[0], dilation[0])
        self.conv_mask = conv_mask
        self.separable = separable
        self.residual_mode = residual_mode
        self.se = se

        inplanes_loop = inplanes
        conv = nn.ModuleList()

        for _ in range(repeat - 1):
            # Stride last means only the last convolution in block will have stride
            if stride_last:
                stride_val = [1]
            else:
                stride_val = stride

            conv.extend(
                self._get_conv_bn_layer(
                    inplanes_loop,
                    planes,
                    kernel_size=kernel_size,
                    stride=stride_val,
                    dilation=dilation,
                    padding=padding_val,
                    groups=groups,
                    heads=heads,
                    separable=separable,
                    normalization=normalization,
                    norm_groups=norm_groups,
                )
            )

            conv.extend(self._get_act_dropout_layer(drop_prob=dropout, activation=activation))

            inplanes_loop = planes

        conv.extend(
            self._get_conv_bn_layer(
                inplanes_loop,
                planes,
                kernel_size=kernel_size,
                stride=stride,
                dilation=dilation,
                padding=padding_val,
                groups=groups,
                heads=heads,
                separable=separable,
                normalization=normalization,
                norm_groups=norm_groups,
            )
        )

        if se:
            conv.append(
                SqueezeExcite(
                    planes,
                    reduction_ratio=se_reduction_ratio,
                    context_window=se_context_window,
                    interpolation_mode=se_interpolation_mode,
                    activation=activation,
                )
            )

        self.mconv = conv

        res_panes = residual_panes.copy()
        self.dense_residual = residual

        if residual:
            res_list = nn.ModuleList()

            if residual_mode == 'stride_add':
                stride_val = stride
            else:
                stride_val = [1]

            if len(residual_panes) == 0:
                res_panes = [inplanes]
                self.dense_residual = False
            for ip in res_panes:
                res = nn.ModuleList(
                    self._get_conv_bn_layer(
                        ip,
                        planes,
                        kernel_size=1,
                        normalization=normalization,
                        norm_groups=norm_groups,
                        stride=stride_val,
                    )
                )

                res_list.append(res)

            self.res = res_list
        else:
            self.res = None

        self.mout = nn.Sequential(*self._get_act_dropout_layer(drop_prob=dropout, activation=activation))

    def _get_conv(
        self,
        in_channels,
        out_channels,
        kernel_size=11,
        stride=1,
        dilation=1,
        padding=0,
        bias=False,
        groups=1,
        heads=-1,
        separable=False,
    ):
        use_mask = self.conv_mask
        if use_mask:
            return MaskedConv1d(
                in_channels,
                out_channels,
                kernel_size,
                stride=stride,
                dilation=dilation,
                padding=padding,
                bias=bias,
                groups=groups,
                heads=heads,
                use_mask=use_mask,
            )
        else:
            return nn.Conv1d(
                in_channels,
                out_channels,
                kernel_size,
                stride=stride,
                dilation=dilation,
                padding=padding,
                bias=bias,
                groups=groups,
            )

    def _get_conv_bn_layer(
        self,
        in_channels,
        out_channels,
        kernel_size=11,
        stride=1,
        dilation=1,
        padding=0,
        bias=False,
        groups=1,
        heads=-1,
        separable=False,
        normalization="batch",
        norm_groups=1,
    ):
        if norm_groups == -1:
            norm_groups = out_channels

        if separable:
            layers = [
                self._get_conv(
                    in_channels,
                    in_channels,
                    kernel_size,
                    stride=stride,
                    dilation=dilation,
                    padding=padding,
                    bias=bias,
                    groups=in_channels,
                    heads=heads,
                ),
                self._get_conv(
                    in_channels,
                    out_channels,
                    kernel_size=1,
                    stride=1,
                    dilation=1,
                    padding=0,
                    bias=bias,
                    groups=groups,
                ),
            ]
        else:
            layers = [
                self._get_conv(
                    in_channels,
                    out_channels,
                    kernel_size,
                    stride=stride,
                    dilation=dilation,
                    padding=padding,
                    bias=bias,
                    groups=groups,
                )
            ]

        if normalization == "group":
            layers.append(nn.GroupNorm(num_groups=norm_groups, num_channels=out_channels))
        elif normalization == "instance":
            layers.append(nn.GroupNorm(num_groups=out_channels, num_channels=out_channels))
        elif normalization == "layer":
            layers.append(nn.GroupNorm(num_groups=1, num_channels=out_channels))
        elif normalization == "batch":
            layers.append(nn.BatchNorm1d(out_channels, eps=1e-3, momentum=0.1))
        else:
            raise ValueError(
                f"Normalization method ({normalization}) does not match" f" one of [batch, layer, group, instance]."
            )

        if groups > 1:
            layers.append(GroupShuffle(groups, out_channels))
        return layers

    def _get_act_dropout_layer(self, drop_prob=0.2, activation=None):
        if activation is None:
            activation = nn.Hardtanh(min_val=0.0, max_val=20.0)
        layers = [activation, nn.Dropout(p=drop_prob)]
        return layers

    def forward(self, input_: Tuple[List[Tensor], Optional[Tensor]]):
        # type: (Tuple[List[Tensor], Optional[Tensor]]) -> Tuple[List[Tensor], Optional[Tensor]] # nopep8
        lens_orig = None
        xs = input_[0]
        if len(input_) == 2:
            xs, lens_orig = input_

        # compute forward convolutions
        out = xs[-1]

        lens = lens_orig
        for i, l in enumerate(self.mconv):
            # if we're doing masked convolutions, we need to pass in and
            # possibly update the sequence lengths
            # if (i % 4) == 0 and self.conv_mask:
            if isinstance(l, MaskedConv1d):
                out, lens = l(out, lens)
            else:
                out = l(out)

        # compute the residuals
        if self.res is not None:
            for i, layer in enumerate(self.res):
                res_out = xs[i]
                for j, res_layer in enumerate(layer):
                    if isinstance(res_layer, MaskedConv1d):
                        res_out, _ = res_layer(res_out, lens_orig)
                    else:
                        res_out = res_layer(res_out)

                if self.residual_mode == 'add' or self.residual_mode == 'stride_add':
                    out = out + res_out
                else:
                    out = torch.max(out, res_out)

        # compute the output
        out = self.mout(out)
        if self.res is not None and self.dense_residual:
            return xs + [out], lens

        return [out], lens


# Register swish activation function
jasper_activations['swish'] = Swish



# Copyright (c) 2019 NVIDIA Corporation
import torch
from torch.utils.data import Dataset

from nemo.collections.asr.parts import collections, parsers
from nemo.collections.asr.parts.segment import AudioSegment


class AudioOnlyDataset(Dataset):
    def __init__(
        self, manifest_filepath, n_segments=0, max_duration=None, min_duration=None, trim=False,
    ):
        """See AudioDataLayer"""
        self.collection = collections.ASRAudioText(
            manifests_files=manifest_filepath.split(','),
            parser=parsers.make_parser(),
            min_duration=min_duration,
            max_duration=max_duration,
        )
        self.trim = trim
        self.n_segments = n_segments

    def AudioCollateFunc(self, batch):
        def find_max_len(seq, index):
            max_len = -1
            for item in seq:
                if item[index].size(0) > max_len:
                    max_len = item[index].size(0)
            return max_len

        batch_size = len(batch)

        audio_signal, audio_lengths = None, None
        if batch[0][0] is not None:
            if self.n_segments > 0:
                max_audio_len = self.n_segments
            else:
                max_audio_len = find_max_len(batch, 0)

            audio_signal = torch.zeros(batch_size, max_audio_len, dtype=torch.float)
            audio_lengths = []
            for i, s in enumerate(batch):
                audio_signal[i].narrow(0, 0, s[0].size(0)).copy_(s[0])
                audio_lengths.append(s[1])
            audio_lengths = torch.tensor(audio_lengths, dtype=torch.long)

        return audio_signal, audio_lengths

    def __getitem__(self, index):
        example = self.collection[index]
        features = AudioSegment.segment_from_file(example.audio_file, n_segments=self.n_segments, trim=self.trim,)
        features = torch.tensor(features.samples, dtype=torch.float)
        f, fl = features, torch.tensor(features.shape[0]).long()

        return f, fl

    def __len__(self):
        return len(self.collection)



# Copyright (c) 2019 NVIDIA Corporation
import torch
import torch.nn as nn

from nemo.backends.pytorch.nm import LossNM
from nemo.core.neural_types import *
from nemo.utils.decorators import add_port_docs


class CTCLossNM(LossNM):
    """
    Neural Module wrapper for pytorch's ctcloss
    Args:
        num_classes (int): Number of characters in ASR model's vocab/labels.
            This count should not include the CTC blank symbol.
        zero_infinity (bool): Whether to zero infinite losses and the associated gradients.
            By default, it is False. Infinite losses mainly occur when the inputs are too
            short to be aligned to the targets.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "log_probs": NeuralType({1: AxisType(TimeTag), 0: AxisType(BatchTag), 2: AxisType(ChannelTag),}),
            # "targets": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag)}),
            # "input_length": NeuralType({0: AxisType(BatchTag)}),
            # "target_length": NeuralType({0: AxisType(BatchTag)}),
            "log_probs": NeuralType(('B', 'T', 'D'), LogprobsType()),
            "targets": NeuralType(('B', 'T'), LabelsType()),
            "input_length": NeuralType(tuple('B'), LengthsType()),
            "target_length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        loss:
            NeuralType(None)
        """
        # return {"loss": NeuralType(None)}
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, num_classes, zero_infinity=False):
        super().__init__()

        self._blank = num_classes
        self._criterion = nn.CTCLoss(blank=self._blank, reduction='none', zero_infinity=zero_infinity)

    def _loss(self, log_probs, targets, input_length, target_length):
        input_length = input_length.long()
        target_length = target_length.long()
        targets = targets.long()
        loss = self._criterion(log_probs.transpose(1, 0), targets, input_length, target_length)
        # note that this is different from reduction = 'mean'
        # because we are not dividing by target lengths
        loss = torch.mean(loss)
        return loss

    def _loss_function(self, **kwargs):
        return self._loss(*(kwargs.values()))



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
Evaluate predictions JSON file, w.r.t. ground truth file.
This file contains code artifacts adapted from the original implementation:
https://github.com/google-research/google-research/blob/master/schema_guided_dst/evaluate.py
"""

import collections
import glob
import json

import numpy as np

import nemo.collections.nlp.data.datasets.sgd_dataset.metrics as metrics
from nemo.utils import logging

__all__ = [
    'get_in_domain_services',
    'get_dataset_as_dict',
    'ALL_SERVICES',
    'SEEN_SERVICES',
    'UNSEEN_SERVICES',
    'get_metrics',
    'PER_FRAME_OUTPUT_FILENAME',
]

ALL_SERVICES = "#ALL_SERVICES"
SEEN_SERVICES = "#SEEN_SERVICES"
UNSEEN_SERVICES = "#UNSEEN_SERVICES"

# Name of the file containing all predictions and their corresponding frame metrics.
PER_FRAME_OUTPUT_FILENAME = "dialogues_and_metrics.json"


def get_service_set(schema_path):
    """Get the set of all services present in a schema."""
    service_set = set()
    with open(schema_path) as f:
        schema = json.load(f)
        for service in schema:
            service_set.add(service["service_name"])
        f.close()
    return service_set


def get_in_domain_services(schema_path, service_set):
    """Get the set of common services between a schema and set of services.
    Args:
        schema_path (str): path to schema file
        service_set (set): set of services
    """
    return get_service_set(schema_path) & service_set


def get_dataset_as_dict(file_path_patterns):
    """Read the DSTC8/SGD json dialogue data as dictionary with dialog ID as keys."""
    dataset_dict = {}
    if isinstance(file_path_patterns, list):
        list_fp = file_path_patterns
    else:
        list_fp = sorted(glob.glob(file_path_patterns))
    for fp in list_fp:
        if PER_FRAME_OUTPUT_FILENAME in fp:
            continue
        logging.debug("Loading file: %s", fp)
        with open(fp) as f:
            data = json.load(f)
            if isinstance(data, list):
                for dial in data:
                    dataset_dict[dial["dialogue_id"]] = dial
            elif isinstance(data, dict):
                dataset_dict.update(data)
            f.close()
    return dataset_dict


def get_metrics(dataset_ref, dataset_hyp, service_schemas, in_domain_services, joint_acc_across_turn, no_fuzzy_match):
    """Calculate the DSTC8/SGD metrics.

  Args:
    dataset_ref: The ground truth dataset represented as a dict mapping dialogue
      id to the corresponding dialogue.
    dataset_hyp: The predictions in the same format as `dataset_ref`.
    service_schemas: A dict mapping service name to the schema for the service.
    in_domain_services: The set of services which are present in the training
      set.
    schemas: Schemas with information for all services

  Returns:
    A dict mapping a metric collection name to a dict containing the values
    for various metrics. Each metric collection aggregates the metrics across
    a specific set of frames in the dialogues.
  """
    # Metrics can be aggregated in various ways, eg over all dialogues, only for
    # dialogues containing unseen services or for dialogues corresponding to a
    # single service. This aggregation is done through metric_collections, which
    # is a dict mapping a collection name to a dict, which maps a metric to a list
    # of values for that metric. Each value in this list is the value taken by
    # the metric on a frame.
    metric_collections = collections.defaultdict(lambda: collections.defaultdict(list))

    # Ensure the dialogs in dataset_hyp also occur in dataset_ref.
    assert set(dataset_hyp.keys()).issubset(set(dataset_ref.keys()))
    logging.debug("len(dataset_hyp)=%d, len(dataset_ref)=%d", len(dataset_hyp), len(dataset_ref))

    # Store metrics for every frame for debugging.
    per_frame_metric = {}

    for dial_id, dial_hyp in dataset_hyp.items():
        seen = True
        dial_ref = dataset_ref[dial_id]

        if set(dial_ref["services"]) != set(dial_hyp["services"]):
            raise ValueError(
                "Set of services present in ground truth and predictions don't match "
                "for dialogue with id {}".format(dial_id)
            )

        joint_metrics = [metrics.JOINT_GOAL_ACCURACY, metrics.JOINT_CAT_ACCURACY, metrics.JOINT_NONCAT_ACCURACY]
        for turn_id, (turn_ref, turn_hyp) in enumerate(zip(dial_ref["turns"], dial_hyp["turns"])):
            metric_collections_per_turn = collections.defaultdict(lambda: collections.defaultdict(lambda: 1.0))
            if turn_ref["speaker"] != turn_hyp["speaker"]:
                raise ValueError("Speakers don't match in dialogue with id {}".format(dial_id))

            # Skip system turns because metrics are only computed for user turns.
            if turn_ref["speaker"] != "USER":
                continue

            if turn_ref["utterance"] != turn_hyp["utterance"]:
                logging.error("Ref utt: %s", turn_ref["utterance"])
                logging.error("Hyp utt: %s", turn_hyp["utterance"])
                raise ValueError("Utterances don't match for dialogue with id {}".format(dial_id))

            hyp_frames_by_service = {frame["service"]: frame for frame in turn_hyp["frames"]}

            # Calculate metrics for each frame in each user turn.
            for frame_ref in turn_ref["frames"]:
                service_name = frame_ref["service"]
                if service_name not in hyp_frames_by_service:
                    raise ValueError(
                        "Frame for service {} not found in dialogue with id {}".format(service_name, dial_id)
                    )
                service = service_schemas[service_name]
                frame_hyp = hyp_frames_by_service[service_name]

                active_intent_acc = metrics.get_active_intent_accuracy(frame_ref, frame_hyp)
                slot_tagging_f1_scores = metrics.get_slot_tagging_f1(
                    frame_ref, frame_hyp, turn_ref["utterance"], service
                )
                requested_slots_f1_scores = metrics.get_requested_slots_f1(frame_ref, frame_hyp)
                goal_accuracy_dict = metrics.get_average_and_joint_goal_accuracy(
                    frame_ref, frame_hyp, service, no_fuzzy_match
                )

                frame_metric = {
                    metrics.ACTIVE_INTENT_ACCURACY: active_intent_acc,
                    metrics.REQUESTED_SLOTS_F1: requested_slots_f1_scores.f1,
                    metrics.REQUESTED_SLOTS_PRECISION: requested_slots_f1_scores.precision,
                    metrics.REQUESTED_SLOTS_RECALL: requested_slots_f1_scores.recall,
                }
                if slot_tagging_f1_scores is not None:
                    frame_metric[metrics.SLOT_TAGGING_F1] = slot_tagging_f1_scores.f1
                    frame_metric[metrics.SLOT_TAGGING_PRECISION] = slot_tagging_f1_scores.precision
                    frame_metric[metrics.SLOT_TAGGING_RECALL] = slot_tagging_f1_scores.recall
                frame_metric.update(goal_accuracy_dict)

                frame_id = "{:s}-{:03d}-{:s}".format(dial_id, turn_id, frame_hyp["service"])
                per_frame_metric[frame_id] = frame_metric
                # Add the frame-level metric result back to dialogues.
                frame_hyp["metrics"] = frame_metric

                # Get the domain name of the service.
                domain_name = frame_hyp["service"].split("_")[0]
                domain_keys = [ALL_SERVICES, frame_hyp["service"], domain_name]
                if frame_hyp["service"] in in_domain_services and seen:
                    domain_keys.append(SEEN_SERVICES)
                else:
                    seen = False
                    domain_keys.append(UNSEEN_SERVICES)
                for domain_key in domain_keys:
                    for metric_key, metric_value in frame_metric.items():
                        if metric_value != metrics.NAN_VAL:
                            if joint_acc_across_turn and metric_key in joint_metrics:
                                metric_collections_per_turn[domain_key][metric_key] *= metric_value
                            else:
                                metric_collections[domain_key][metric_key].append(metric_value)
            if joint_acc_across_turn:
                # Conduct multiwoz style evaluation that computes joint goal accuracy
                # across all the slot values of all the domains for each turn.
                for domain_key in metric_collections_per_turn:
                    for metric_key, metric_value in metric_collections_per_turn[domain_key].items():
                        metric_collections[domain_key][metric_key].append(metric_value)

    all_metric_aggregate = {}
    for domain_key, domain_metric_vals in metric_collections.items():
        domain_metric_aggregate = {}
        for metric_key, value_list in domain_metric_vals.items():
            if value_list:
                # Metrics are macro-averaged across all frames.
                domain_metric_aggregate[metric_key] = round(float(np.mean(value_list)) * 100.0, 2)
            else:
                domain_metric_aggregate[metric_key] = metrics.NAN_VAL
        all_metric_aggregate[domain_key] = domain_metric_aggregate
    return all_metric_aggregate, per_frame_metric



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.data.datasets.glue_benchmark_dataset.glue_benchmark_dataset import GLUEDataset
from nemo.collections.nlp.data.datasets.joint_intent_slot_dataset.joint_intent_slot_dataset import (
    BertJointIntentSlotDataset,
    BertJointIntentSlotInferDataset,
)
from nemo.collections.nlp.data.datasets.lasertagger_dataset import LaserTaggerDataset
from nemo.collections.nlp.data.datasets.lm_bert_dataset import (
    BertPretrainingDataset,
    BertPretrainingPreprocessedDataset,
)
from nemo.collections.nlp.data.datasets.lm_transformer_dataset import LanguageModelingDataset
from nemo.collections.nlp.data.datasets.machine_translation_dataset import TranslationDataset
from nemo.collections.nlp.data.datasets.multiwoz_dataset.multiwoz_dataset import MultiWOZDataDesc, MultiWOZDataset
from nemo.collections.nlp.data.datasets.punctuation_capitalization_dataset import (
    BertPunctuationCapitalizationDataset,
    BertPunctuationCapitalizationInferDataset,
)
from nemo.collections.nlp.data.datasets.qa_squad_dataset.qa_squad_dataset import SquadDataset
from nemo.collections.nlp.data.datasets.sgd_dataset.schema_embedding_dataset import SchemaEmbeddingDataset
from nemo.collections.nlp.data.datasets.sgd_dataset.sgd_dataset import SGDDataset
from nemo.collections.nlp.data.datasets.text_classification import (
    BertTextClassificationDataset,
    TextClassificationDataDesc,
)
from nemo.collections.nlp.data.datasets.token_classification_dataset import (
    BertTokenClassificationDataset,
    BertTokenClassificationInferDataset,
)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.data import SquadDataset
from nemo.collections.nlp.nm.data_layers.text_datalayer import TextDataLayer
from nemo.core import ChannelType, LabelsType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['BertQuestionAnsweringDataLayer']


class BertQuestionAnsweringDataLayer(TextDataLayer):
    """
    Creates the data layer to use for Question Answering classification task.

    Args:
        data_file (str): data_file in *.json.
        tokenizer (obj): Tokenizer object, e.g. NemoBertTokenizer.
        version_2_with_negative (bool): True if training should allow
            unanswerable questions.
        doc_stride (int): When splitting up a long document into chunks,
            how much stride to take between chunks.
        max_query_length (int): All training files which have a duration less
            than min_duration are dropped. Can't be used if the `utt2dur` file
            does not exist. Defaults to None.
        max_seq_length (int): All training files which have a duration more
            than max_duration are dropped. Can't be used if the `utt2dur` file
            does not exist. Defaults to None.
        mode (str): Use "train", "eval", or "test" to define between
            training and evaluation and inference.
        batch_size (int): Batch size. Defaults to 64.
        dataset_type (Dataset): Question Answering class.
            Defaults to SquadDataset.
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        input_ids:
            indices of tokens which constitute batches of masked text segments
        input_type_ids:
            tensor with 0's and 1's to denote the text segment type
        input_mask:
            bool tensor with 0s in place of tokens to be masked
        start_positions: indices of tokens which constitute start position of answer
        end_positions: indices of tokens which constitute end position of answer
        unique_ids: id of the Question answer example this instance belongs to
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_type_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_mask": NeuralType(('B', 'T'), ChannelType()),
            "unique_ids": NeuralType(tuple('B'), ChannelType()),
            "start_positions": NeuralType(tuple('B'), ChannelType(), optional=True),
            "end_positions": NeuralType(tuple('B'), ChannelType(), optional=True),
        }

    def __init__(
        self,
        data_file,
        tokenizer,
        version_2_with_negative,
        doc_stride,
        max_query_length,
        max_seq_length,
        mode,
        batch_size=64,
        use_cache=True,
        shuffle=False,
        dataset_type=SquadDataset,
    ):
        dataset_params = {
            'data_file': data_file,
            'mode': mode,
            'tokenizer': tokenizer,
            'version_2_with_negative': version_2_with_negative,
            'max_query_length': max_query_length,
            'max_seq_length': max_seq_length,
            'use_cache': use_cache,
            'doc_stride': doc_stride,
        }

        super().__init__(dataset_type, dataset_params, batch_size, shuffle=shuffle)



# -*- coding: utf-8 -*-
# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
# -*- coding: utf-8 -*-
#
# Copyright (C) IBM Corporation 2019
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

__author__ = "Tomasz Kornuta"

"""
This file contains code artifacts adapted from the original implementation:
https://github.com/IBM/pytorchpipe/blob/develop/ptp/components/models/vision/image_encoder.py
"""

from typing import Optional

import torch
import torchvision.models as models

from nemo.backends.pytorch.nm import TrainableNM
from nemo.core.neural_types import AxisKind, AxisType, ImageFeatureValue, ImageValue, LogitsType, NeuralType
from nemo.utils import logging
from nemo.utils.configuration_parsing import get_value_from_dictionary
from nemo.utils.decorators import add_port_docs

__all__ = ['ImageEncoder']


class ImageEncoder(TrainableNM):
    """
    Neural Module implementing a general-usage image encoderds.
    It encapsulates several models from TorchVision (VGG16, ResNet152 and DensNet121, naming a few).
    Offers two operation modes and can return: image embeddings vs feature maps.
    """

    def __init__(
        self,
        model_type: str,
        output_size: Optional[int] = None,
        return_feature_maps: bool = False,
        pretrained: bool = False,
        name: Optional[str] = None,
    ):
        """
        Initializes the ``ImageEncoder`` model, creates the required "backbone".

        Args:
            model_type: Type of backbone (Options: VGG16 | DenseNet121 | ResNet152 | ResNet50)
            output_size: Size of the output layer (Optional, Default: None)
            return_feature_maps: Return mode: image embeddings vs feature maps (Default: False)
            pretrained: Loads pretrained model (Default: False)
            name: Name of the module (DEFAULT: None)
        """
        TrainableNM.__init__(self, name=name)

        # Get operation modes.
        self._return_feature_maps = return_feature_maps

        # Get model type.
        self._model_type = get_value_from_dictionary(
            model_type, "vgg16 | densenet121 | resnet152 | resnet50".split(" | ")
        )

        # Get output size (optional - not in feature_maps).
        self._output_size = output_size

        if self._model_type == 'vgg16':
            # Get VGG16
            self._model = models.vgg16(pretrained=pretrained)

            if self._return_feature_maps:
                # Use only the "feature encoder".
                self._model = self._model.features

                # Remember the output feature map dims.
                self._feature_map_height = 7
                self._feature_map_width = 7
                self._feature_map_depth = 512

            else:
                # Use the whole model, but "reshape"/reinstantiate the last layer ("FC6").
                self._model.classifier._modules['6'] = torch.nn.Linear(4096, self._output_size)

        elif self._model_type == 'densenet121':
            # Get densenet121
            self._model = models.densenet121(pretrained=pretrained)

            if self._return_feature_maps:
                raise ConfigurationError("'densenet121' doesn't support 'return_feature_maps' mode (yet)")

            # Use the whole model, but "reshape"/reinstantiate the last layer ("FC6").
            self._model.classifier = torch.nn.Linear(1024, self._output_size)

        elif self._model_type == 'resnet152':
            # Get resnet152
            self._model = models.resnet152(pretrained=pretrained)

            if self._return_feature_maps:
                # Get all modules exluding last (avgpool) and (fc)
                modules = list(self._model.children())[:-2]
                self._model = torch.nn.Sequential(*modules)

                # Remember the output feature map dims.
                self._feature_map_height = 7
                self._feature_map_width = 7
                self._feature_map_depth = 2048

            else:
                # Use the whole model, but "reshape"/reinstantiate the last layer ("FC6").
                self._model.fc = torch.nn.Linear(2048, self._output_size)

        elif self._model_type == 'resnet50':
            # Get resnet50
            self._model = models.resnet50(pretrained=pretrained)

            if self._return_feature_maps:
                # Get all modules exluding last (avgpool) and (fc)
                modules = list(self._model.children())[:-2]
                self._model = torch.nn.Sequential(*modules)

                # Remember the output feature map dims.
                self._feature_map_height = 7
                self._feature_map_width = 7
                self._feature_map_depth = 2048

            else:
                # Use the whole model, but "reshape"/reinstantiate the last layer ("FC6").
                self._model.fc = torch.nn.Linear(2048, self._output_size)

    @property
    @add_port_docs()
    def input_ports(self):
        """
        Returns definitions of module input ports.
        """
        return {
            "inputs": NeuralType(
                axes=(
                    AxisType(kind=AxisKind.Batch),
                    AxisType(kind=AxisKind.Channel, size=3),
                    AxisType(kind=AxisKind.Height, size=224),
                    AxisType(kind=AxisKind.Width, size=224),
                ),
                elements_type=ImageValue(),
                # TODO: actually encoders pretrained on ImageNet require special image normalization.
                # Probably this should be a new image type.
            )
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """
        Returns definitions of module output ports.
        """
        # Return neural type.
        if self._return_feature_maps:
            return {
                "outputs": NeuralType(
                    axes=(
                        AxisType(kind=AxisKind.Batch),
                        AxisType(kind=AxisKind.Channel, size=self._feature_map_depth),
                        AxisType(kind=AxisKind.Height, size=self._feature_map_height),
                        AxisType(kind=AxisKind.Width, size=self._feature_map_width),
                    ),
                    elements_type=ImageFeatureValue(),
                )
            }
        else:
            return {
                "outputs": NeuralType(
                    axes=(AxisType(kind=AxisKind.Batch), AxisType(kind=AxisKind.Any, size=self._output_size),),
                    elements_type=LogitsType(),
                )
            }

    def forward(self, inputs):
        """
        Main forward pass of the model.

        Args:
            inputs: expected stream containing images [BATCH_SIZE x IMAGE_DEPTH x IMAGE_HEIGHT x IMAGE WIDTH]
        
        Returns:
            outpus: added stream containing outputs [BATCH_SIZE x OUTPUT_SIZE]
                OR [BATCH_SIZE x OUTPUT_DEPTH x OUTPUT_HEIGHT x OUTPUT_WIDTH]
        """
        # print("{}: input shape: {}, device: {}\n".format(self.name, inputs.shape, inputs.device))

        outputs = self._model(inputs)

        # Add outputs to datadict.
        return outputs



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""Pytorch Dataset for training Neural Machine Translation."""
import os
import re

import numpy as np
from torch.utils.data import Dataset

from nemo import logging
from nemo.collections.nlp.data.datasets.datasets_utils import dataset_to_ids, if_exist

__all__ = ['LanguageModelingDataset', 'LanguageModelDataDesc']


class LanguageModelingDataset(Dataset):
    def __init__(self, tokenizer, dataset, max_seq_length=512, batch_step=None):
        self.tokenizer = tokenizer
        self.max_seq_length = max_seq_length
        self.batch_step = batch_step or self.max_seq_length
        ids = dataset_to_ids(dataset, tokenizer, add_bos_eos=False)
        self.ids = np.array([j for i in ids for j in i])

    def __len__(self):
        return (len(self.ids) - self.max_seq_length) // self.batch_step

    def __getitem__(self, idx):
        left = idx * self.batch_step
        right = left + self.max_seq_length
        src_ids = self.ids[left:right]
        labels = self.ids[left + 1 : right + 1]
        src_mask = (src_ids != self.tokenizer.pad_id).astype(np.float32)
        return src_ids, src_mask, labels


class LanguageModelDataDesc:
    def __init__(self, dataset_name, data_dir, do_lower_case):
        if dataset_name == 'wikitext-2':
            if not os.path.exists(data_dir):
                raise FileNotFoundError(
                    "Dataset not found. Run 'get_wkt2.sh DATA_DIR' from examples/nlp/language_modeling"
                )
            self.vocab_size = self.create_vocab_lm(data_dir, do_lower_case)
            self.data_dir = data_dir
        else:
            logging.warning(
                "Looks like you passed a dataset name that isn't "
                "already supported by NeMo. Please make sure that "
                "you build the preprocessing method for it."
            )

    def create_vocab_lm(self, data_dir, do_lower_case):
        if if_exist(data_dir, ['train.txt', 'vocab.txt']):
            logging.info("Vocabulary has been created.")
            with open(os.path.join(data_dir, 'vocab.txt'), 'r') as f:
                vocab_size = len(f.readlines())
            return vocab_size

        logging.info(f'Creating vocabulary from training data at {data_dir}')

        with open(f'{data_dir}/train.txt', 'r') as f:
            txt = f.read()
        if do_lower_case:
            txt = txt.lower()
        lines = re.split(r'[\n]', txt)
        sentences = [line.strip().split() for line in lines if line.strip()]

        vocab = {"[PAD]": 0, "[SEP]": 1, "[CLS]": 2, "[MASK]": 3}
        idx = 4
        for sentence in sentences:
            for word in sentence:
                if word not in vocab:
                    vocab[word] = idx
                    idx += 1

        with open(f'{data_dir}/vocab.txt', 'w') as f:
            for word in sorted(vocab.keys()):
                f.write(word + '\n')
        logging.info(f"Created vocabulary of size {len(vocab)}")

        return len(vocab)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import csv

__all__ = ['DataProcessor']


class DataProcessor(object):
    """Base class for data converters for sequence classification data sets."""

    def get_train_examples(self, data_dir):
        """Gets a collection of `InputExample`s for the train set."""
        raise NotImplementedError()

    def get_dev_examples(self, data_dir):
        """Gets a collection of `InputExample`s for the dev set."""
        raise NotImplementedError()

    def get_labels(self):
        """Gets the list of labels for this data set."""
        raise NotImplementedError()

    @classmethod
    def _read_tsv(cls, input_file, quotechar=None):
        """Reads a tab separated value file."""
        with open(input_file, "r", encoding="utf-8-sig") as f:
            reader = csv.reader(f, delimiter="\t", quotechar=quotechar)
            lines = []
            for line in reader:
                # if sys.version_info[0] == 2:
                #     line = list(unicode(cell, 'utf-8') for cell in line)
                lines.append(line)
            return lines



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import torch

from nemo.backends.pytorch import LossNM
from nemo.collections.nlp.utils.data_utils import mask_padded_tokens
from nemo.core import LabelsType, LogitsType, LossType, MaskType, NeuralType

__all__ = ['SmoothedCrossEntropyLoss']


class SmoothedCrossEntropyLoss(LossNM):
    """
    Neural module which calculates CrossEntropyLoss and
    1) excludes padding tokens from loss calculation
    2) allows to use label smoothing regularization
    3) allows to calculate loss for the desired number of last tokens

    Args:
        label_smoothing (float): label smoothing regularization coefficient
        predict_last_k (int): how many last tokens to use for the loss
            calculation, important for fast evaluation of LM perplexity
    """

    @property
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            "logits": NeuralType(('B', 'T', 'D'), LogitsType()),
            "labels": NeuralType(('B', 'T'), LabelsType()),
            "output_mask": NeuralType(('B', 'T'), MaskType(), optional=True),
        }

    @property
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        # return {"loss": NeuralType(None)}
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, pad_id=None, label_smoothing=0, predict_last_k=0):
        LossNM.__init__(self)

        self._loss_fn = SmoothedCrossEntropy(label_smoothing, predict_last_k)
        self._pad_id = pad_id

    def _loss_function(self, logits, labels, output_mask=None):
        if output_mask is not None:
            labels_mask = output_mask
        elif self._pad_id is not None:
            labels_mask = mask_padded_tokens(labels, self._pad_id).to(logits.dtype)
        else:
            raise ValueError("Both output_mask and pad_id are None")

        if labels_mask.dtype is not logits.dtype:
            labels_mask = labels_mask.to(logits.dtype)

        loss = self._loss_fn(logits, labels, labels_mask)
        return loss


class SmoothedCrossEntropy(torch.nn.Module):
    """
    Cross-entropy loss with label smoothing for a batch of sequences.

    Args:
        label_smoothing (float): label smoothing coefficient, usually set
                                 between 0.0 and 0.1 in language modeling
                                 and translation pipelines
        predict_last_k (int): int parameter which sets the number of last
                              tokens to calculate the loss for, for example
            0: (default) calculate loss on the entire sequence (e.g., NMT)
            1: calculate loss on the last token only (e.g., LM evaluation)
            Intermediate values allow to control the trade-off between eval
            time (proportional to the number of batches) and eval performance
            (proportional to the number of context tokens).
    """

    def __init__(self, label_smoothing=0.0, predict_last_k=0):
        super().__init__()
        self._smoothing = label_smoothing
        self._predict_last_k = predict_last_k

    def forward(self, logits, labels, output_mask, eps=1e-6):
        """
        Args:
            logits: float tensor of shape batch_size x seq_len x vocab_size, values should be log probabilities
            labels: int tensor of shape batch_size x seq_len
            output_mask: binary tensor of shape batch_size x seq_len
            eps: epsilon param to avoid divide by zero in loss calculation
        """
        batch_size, seq_len, vocab_size = logits.size()
        smoothing = vocab_size * self._smoothing / (vocab_size - 1)
        target_logits = logits.gather(2, labels.unsqueeze(2)).squeeze(2)
        smoothing_logits = logits.mean(dim=-1)
        neg_log_likelihood = (1.0 - smoothing) * target_logits + smoothing * smoothing_logits
        neg_log_likelihood = neg_log_likelihood[:, -self._predict_last_k :]
        output_mask = output_mask[:, -self._predict_last_k :]
        neg_log_likelihood = -torch.sum(neg_log_likelihood * output_mask)
        neg_log_likelihood = neg_log_likelihood / (output_mask.sum() + eps)
        return neg_log_likelihood



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from abc import ABC, abstractmethod
from typing import List

__all__ = ['TokenizerSpec']


class TokenizerSpec(ABC):
    @abstractmethod
    def text_to_tokens(self, text):
        pass

    @abstractmethod
    def tokens_to_text(self, tokens):
        pass

    @abstractmethod
    def tokens_to_ids(self, tokens):
        pass

    @abstractmethod
    def ids_to_tokens(self, ids):
        pass

    @abstractmethod
    def text_to_ids(self, text):
        pass

    @abstractmethod
    def ids_to_text(self, ids):
        pass

    def add_special_tokens(self, special_tokens: List[str]):
        raise NotImplementedError("To be implemented")



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import os

from transformers import AlbertTokenizer, BertTokenizer, RobertaTokenizer

import nemo
from nemo.utils import logging

try:
    __megatron_utils_satisfied = True
    from nemo.collections.nlp.nm.trainables.common.megatron.megatron_utils import (
        get_megatron_vocab_file,
        is_lower_cased_megatron,
    )

except Exception as e:
    logging.error('Failed to import Megatron utils: `{}` ({})'.format(str(e), type(e)))
    __megatron_utils_satisfied = False


__all__ = ['MODEL_SPECIAL_TOKENS', 'TOKENIZERS', 'get_tokenizer', 'get_bert_special_tokens']

MODEL_SPECIAL_TOKENS = {
    'bert': {
        'unk_token': '[UNK]',
        'sep_token': '[SEP]',
        'pad_token': '[PAD]',
        'bos_token': '[CLS]',
        'mask_token': '[MASK]',
        'eos_token': '[SEP]',
        'cls_token': '[CLS]',
    },
    'roberta': {
        'unk_token': '<unk>',
        'sep_token': '</s>',
        'pad_token': '<pad>',
        'bos_token': '<s>',
        'mask_token': '<mask>',
        'eos_token': '</s>',
        'cls_token': '<s>',
    },
    'albert': {
        'unk_token': '<unk>',
        'sep_token': '[SEP]',
        'pad_token': '<pad>',
        'bos_token': '[CLS]',
        'mask_token': '[MASK]',
        'eos_token': '[SEP]',
        'cls_token': '[CLS]',
    },
}


TOKENIZERS = {'bert': BertTokenizer, 'albert': AlbertTokenizer, 'roberta': RobertaTokenizer}


def get_bert_special_tokens(bert_derivative):
    return MODEL_SPECIAL_TOKENS[bert_derivative]


def get_tokenizer(
    tokenizer_name,
    pretrained_model_name,
    tokenizer_model=None,
    special_tokens=None,
    vocab_file=None,
    do_lower_case=False,
):
    '''
    Args:
    tokenizer_name: sentencepiece or nemobert
    pretrained_mode_name ('str'): name of the pretrained model from the hugging face list or 'megatron',
        for example: bert-base-cased
        To see the list of pretrained models, use: nemo_nlp.nm.trainables.get_bert_models_list()
    tokenizer_model (path): only used for sentencepiece tokenizer
    special_tokens (dict): dict of special tokens (Optional)
    vocab_file (str): path to vocab file
    do_lower_case (bool): (whether to apply lower cased) - only applicable when tokenizer is build with vocab file
    '''
    # Check if we can use Megatron utils.
    if __megatron_utils_satisfied:
        if 'megatron' in pretrained_model_name:
            do_lower_case = is_lower_cased_megatron(pretrained_model_name)
            vocab_file = get_megatron_vocab_file(pretrained_model_name)
            return nemo.collections.nlp.data.tokenizers.NemoBertTokenizer(
                vocab_file=vocab_file, do_lower_case=do_lower_case
            )

    if tokenizer_name == 'nemobert':
        tokenizer = nemo.collections.nlp.data.tokenizers.NemoBertTokenizer(
            pretrained_model=pretrained_model_name, vocab_file=vocab_file, do_lower_case=do_lower_case
        )
    elif tokenizer_name == 'sentencepiece':
        if not os.path.exists(tokenizer_model):
            raise FileNotFoundError(f'{tokenizer_model} tokenizer model not found')

        tokenizer = nemo.collections.nlp.data.tokenizers.SentencePieceTokenizer(model_path=tokenizer_model)
        model_type = pretrained_model_name.split('-')[0]
        if special_tokens is None:
            if model_type not in MODEL_SPECIAL_TOKENS:
                logging.info(f'No special tokens found for {model_type}.')
            else:
                special_tokens = MODEL_SPECIAL_TOKENS[model_type]
        tokenizer.add_special_tokens(special_tokens)
    else:
        raise ValueError(f'{tokenizer_name} is not supported')
    return tokenizer



# =============================================================================
# Copyright (c) 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from typing import Optional

import torch

from nemo.backends.pytorch.nm import LossNM
from nemo.core.neural_types import ClassificationTarget, LogprobsType, LossType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['NLLLoss']


class NLLLoss(LossNM):
    """ Class representing a simple NLL loss. """

    def __init__(self, name: Optional[str] = None):
        """
        Constructor.

        Args:
            name: Name of the module (DEFAULT: None)
        """
        # Call the base class constructor.
        LossNM.__init__(self, name=name)
        # Set criterion.
        self._criterion = torch.nn.NLLLoss()

    @property
    @add_port_docs()
    def input_ports(self):
        """ Returns definitions of module input ports. """
        return {
            "predictions": NeuralType(axes=('B', 'ANY'), elements_type=LogprobsType()),
            "targets": NeuralType(axes=('B'), elements_type=ClassificationTarget()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """ Returns definitions of module output ports. """
        return {"loss": NeuralType(elements_type=LossType())}

    # You need to implement this function
    def _loss_function(self, **kwargs):
        return self._criterion(*(kwargs.values()))



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2018 The Google AI Language Team Authors and The HugginFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import copy

import torch
import torch.nn as nn

from nemo.collections.nlp.nm.trainables.common.transformer.transformer_modules import (
    MultiHeadAttention,
    PositionWiseFF,
)
from nemo.collections.nlp.utils.transformer_utils import form_attention_mask

__all__ = []


class TransformerDecoderBlock(nn.Module):
    """
    Building block of Transformer decoder.

    Args:
        hidden_size: size of the embeddings in the model, also known as d_model
        inner_size: number of neurons in the intermediate part of feed-forward
            net, usually is (4-8 x hidden_size) in the papers
        num_attention_heads: number of heads in multi-head attention
        attn_score_dropout: probability of dropout applied to attention scores
        attn_layer_dropout: probability of dropout applied to the output of the
            attention layers, but before layer normalization
        ffn_dropout: probability of dropout applied to FFN output
        hidden_act: activation function used between two linear layers in FFN
    """

    def __init__(
        self,
        hidden_size,
        inner_size,
        num_attention_heads=1,
        attn_score_dropout=0,
        attn_layer_dropout=0,
        ffn_dropout=0,
        hidden_act="relu",
        use_full_attention=True,
    ):
        super().__init__()

        self.first_sub_layer = MultiHeadAttention(
            hidden_size, num_attention_heads, attn_score_dropout, attn_layer_dropout
        )
        self.second_sub_layer = MultiHeadAttention(
            hidden_size, num_attention_heads, attn_score_dropout, attn_layer_dropout
        )
        self.third_sub_layer = PositionWiseFF(hidden_size, inner_size, ffn_dropout, hidden_act)
        self.use_full_attention = use_full_attention

    def forward(self, decoder_query, decoder_mask, decoder_keys, encoder_states, encoder_mask):
        self_attn_output = self.first_sub_layer(decoder_query, decoder_keys, decoder_keys, decoder_mask)
        enc_dec_attn_output = (
            self.second_sub_layer(self_attn_output, encoder_states, encoder_states, encoder_mask)
            if self.use_full_attention
            else self_attn_output
        )
        output_states = self.third_sub_layer(enc_dec_attn_output)
        return output_states


class TransformerDecoder(nn.Module):
    def __init__(self, num_layers, hidden_size, **kwargs):
        super().__init__()

        layer = TransformerDecoderBlock(hidden_size, **kwargs)
        self.layers = nn.ModuleList([copy.deepcopy(layer) for _ in range(num_layers)])

    def _get_memory_states(self, decoder_states, decoder_mems_list=None, i=0):
        if decoder_mems_list is not None:
            memory_states = torch.cat((decoder_mems_list[i], decoder_states), dim=1)
        else:
            memory_states = decoder_states
        return memory_states

    def forward(
        self, decoder_states, decoder_mask, encoder_states, encoder_mask, decoder_mems_list=None, return_mems=False
    ):
        """
        Args:
            decoder_states: output of the embedding layer (B x L_dec x H)
            decoder_mask: decoder inputs mask (B x L_dec)
            encoder_states: output of the encoder (B x L_enc x H)
            encoder_mask: encoder inputs mask (B x L_enc)
            decoder_mems_list: list of the cached decoder hidden states
                for fast autoregressive generation which will be used instead
                of decoder_states as keys and values if not None
            return_mems: bool, whether to return outputs of all decoder layers
                or the last layer only
        """

        decoder_attn_mask = form_attention_mask(decoder_mask, diagonal=0)
        encoder_attn_mask = form_attention_mask(encoder_mask)

        memory_states = self._get_memory_states(decoder_states, decoder_mems_list, 0)
        cached_mems_list = [memory_states]

        for i, layer in enumerate(self.layers):
            decoder_states = layer(decoder_states, decoder_attn_mask, memory_states, encoder_states, encoder_attn_mask)
            memory_states = self._get_memory_states(decoder_states, decoder_mems_list, i + 1)
            cached_mems_list.append(memory_states)

        if return_mems:
            return cached_mems_list
        else:
            return cached_mems_list[-1]



# Copyright (c) 2019 NVIDIA Corporation
"""A collection of Neural Modules to be used for training a WGAN-GP on MNIST"""
import torch
from torch.utils.data import Dataset
from torchvision import datasets, transforms

from nemo.backends.pytorch.nm import DataLayerNM, LossNM, TrainableNM
from nemo.core import DeviceType
from nemo.core.neural_types import ChannelType, LabelsType, LossType, NeuralType
from nemo.utils.decorators import add_port_docs


class SimpleDiscriminator(TrainableNM):
    """Simple convolutional discrimnator that takes in a 28x28 greyscale image
    and assigns a score to it.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "image": NeuralType(
            #     {
            #         0: AxisType(BatchTag),
            #         1: AxisType(ChannelTag),
            #         2: AxisType(HeightTag, 28),
            #         3: AxisType(WidthTag, 28),
            #     }
            # )
            "image": NeuralType(('B', 'C', 'H', 'W'), ChannelType())
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        # return {"decision": NeuralType({0: AxisType(BatchTag), 1: AxisType(ChannelTag, 1)})}
        return {"decision": NeuralType(('B', 'C'), ChannelType())}

    def __init__(self):
        super().__init__()
        self.layers = torch.nn.Sequential(
            torch.nn.Conv2d(1, 64, 3, padding=1),
            torch.nn.ReLU(),
            torch.nn.Conv2d(64, 128, 3, stride=2, padding=1),
            torch.nn.ReLU(),
            torch.nn.Conv2d(128, 128, 3, stride=2, padding=1),
            torch.nn.ReLU(),
            torch.nn.Conv2d(128, 256, 3, stride=2, padding=1),
            torch.nn.ReLU(),
        )
        self.fc_layer = torch.nn.Linear(256 * 4 * 4, 1)
        self.to(self._device)

    def forward(self, image):
        decision = self.layers(image)
        decision = decision.view(-1, 256 * 4 * 4)
        decision = self.fc_layer(decision)
        return decision


class SimpleGenerator(TrainableNM):
    """Simple convolutional generator that takes a random variable of size
    (64, 4, 4) and produces a 28x28 greyscale image.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "latents": NeuralType(
            #     {
            #         0: AxisType(BatchTag),
            #         1: AxisType(ChannelTag, 64),
            #         2: AxisType(HeightTag, 4),
            #         3: AxisType(WidthTag, 4),
            #     }
            # )
            "latents": NeuralType(('B', 'C', 'H', 'W'), ChannelType())
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "image": NeuralType(
            #     {
            #         0: AxisType(BatchTag),
            #         1: AxisType(ChannelTag),
            #         2: AxisType(HeightTag, 28),
            #         3: AxisType(WidthTag, 28),
            #     }
            # )
            "image": NeuralType(('B', 'C', 'H', 'W'), ChannelType())
        }

    def __init__(self, batch_size):
        super().__init__()
        self._batch_size = batch_size

        self.layers = torch.nn.Sequential(
            torch.nn.ConvTranspose2d(64, 128, 3, stride=2),
            torch.nn.ReLU(),
            torch.nn.ConvTranspose2d(128, 128, 3, stride=2),
            torch.nn.ReLU(),
            torch.nn.ConvTranspose2d(128, 128, 3, stride=2),
            torch.nn.ReLU(),
            torch.nn.Conv2d(128, 1, 12),
            torch.nn.Sigmoid(),
        )
        self.to(self._device)

    def forward(self, latents):
        image = latents
        for layer in self.layers:
            image = layer(image)
        return image


class DiscriminatorLoss(LossNM):
    """Computes the loss from a disciminator score by simply taking the mean
    of all scores in a batch.

    Args:
        neg (bool): Whether to negate the final loss
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.

        decision:
            0: AxisType(BatchTag)

            1: AxisType(ChannelTag, 1)
        """
        return {
            # "decision": NeuralType({0: AxisType(BatchTag), 1: AxisType(ChannelTag, 1)}),
            "decision": NeuralType(('B', 'D'), ChannelType())
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, neg=False):
        super().__init__()
        self.neg = neg

    def _loss(self, decision):
        if self.neg:
            return -torch.mean(decision)
        return torch.mean(decision)

    def _loss_function(self, **kwargs):
        return self._loss(*(kwargs.values()))


class GradientPenalty(LossNM):
    """Compute the gradient penalty of the disciminator

    Args:
        lambda_ (float): lambda parameter indicating the weight of the loss.
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "interpolated_image": NeuralType(
            #     {
            #         0: AxisType(BatchTag),
            #         1: AxisType(ChannelTag),
            #         2: AxisType(HeightTag, 28),
            #         3: AxisType(WidthTag, 28),
            #     }
            # ),
            # "interpolated_decision": NeuralType({0: AxisType(BatchTag), 1: AxisType(ChannelTag, 1)}),
            "interpolated_image": NeuralType(('B', 'C', 'H', 'W'), ChannelType()),
            "interpolated_decision": NeuralType(('B', 'C'), ChannelType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        loss:
            NeuralType(None)
        """
        return {"loss": NeuralType(elements_type=LossType())}

    def __init__(self, lambda_):
        super().__init__()
        self.lambda_ = lambda_

    def _loss(self, interpolated_image, interpolated_decision):
        grad_outputs = torch.ones(interpolated_decision.size(), dtype=interpolated_image.dtype)
        if self.placement != DeviceType.CPU:
            grad_outputs = grad_outputs.cuda()

        gradients = torch.autograd.grad(
            outputs=interpolated_decision,
            inputs=interpolated_image,
            grad_outputs=grad_outputs,
            create_graph=True,
            retain_graph=True,
            only_inputs=True,
        )[0]
        gradients = gradients.view(gradients.size(0), -1)

        gradient_penalty = ((gradients.norm(2, dim=1) - 1) ** 2).mean()
        return self.lambda_ * gradient_penalty

    def _loss_function(self, **kwargs):
        return self._loss(**kwargs)


class InterpolateImage(TrainableNM):
    """Linearly interpolates an image between image1 and image2
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "image1": NeuralType(
            #     {
            #         0: AxisType(BatchTag),
            #         1: AxisType(ChannelTag),
            #         2: AxisType(HeightTag, 28),
            #         3: AxisType(WidthTag, 28),
            #     }
            # ),
            # "image2": NeuralType(
            #     {
            #         0: AxisType(BatchTag),
            #         1: AxisType(ChannelTag),
            #         2: AxisType(HeightTag, 28),
            #         3: AxisType(WidthTag, 28),
            #     }
            # ),
            "image1": NeuralType(('B', 'C', 'H', 'W'), ChannelType()),
            "image2": NeuralType(('B', 'C', 'H', 'W'), ChannelType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "interpolated_image": NeuralType(
            #     {
            #         0: AxisType(BatchTag),
            #         1: AxisType(ChannelTag),
            #         2: AxisType(HeightTag, 28),
            #         3: AxisType(WidthTag, 28),
            #     }
            # )
            "interpolated_image": NeuralType(('B', 'C', 'H', 'W'), ChannelType())
        }

    def __init__(self):
        super().__init__()

    def forward(self, image1, image2):
        alpha = torch.rand(image1.shape[0], 1).unsqueeze(-1).unsqueeze(-1)
        alpha = alpha.to(self._device)
        alpha = alpha.type(image2.dtype)
        image1 = image1.type(image2.dtype)

        interpolated_image = alpha * image1 + ((1 - alpha) * image2)
        return torch.autograd.Variable(interpolated_image, requires_grad=True)


class RandomDataLayer(DataLayerNM):
    """Dummy data layer for return random variables to be used in the generator

    Args:
        batch_size (int)
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        latent:
            0: AxisType(BatchTag)

            1: AxisType(ChannelTag, 64)

            2: AxisType(HeightTag, 4)

            3: AxisType(WidthTag, 4)
        """
        return {
            # "latent": NeuralType(
            #     {
            #         0: AxisType(BatchTag),
            #         1: AxisType(ChannelTag, 64),
            #         2: AxisType(HeightTag, 4),
            #         3: AxisType(WidthTag, 4),
            #     }
            # )
            "latent": NeuralType(('B', 'C', 'H', 'W'), ChannelType())
        }

    def __init__(self, batch_size):
        super().__init__()
        self._batch_size = batch_size

        class DummyDataset(torch.utils.data.Dataset):
            def __init__(self, batch_size):
                super().__init__()
                self._batch_size = batch_size

            def __getitem__(self, i):
                return torch.randn(64, 4, 4)

            def __len__(self):
                return self._batch_size * 2

        self._dataset = DummyDataset(batch_size)

    def __len__(self):
        return self._dataset.__len__()

    @property
    def dataset(self):
        return self._dataset

    @property
    def data_iterator(self):
        return None


class MnistGanDataLayer(DataLayerNM):
    """Wrapper around torchvision's MNIST dataset. Additionally, it returns a
    random variable to be used in the generator.

    Args:
        batch_size (int)
        root (str): Where to store the dataset
        train (bool)
        shuffle (bool)
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "latent": NeuralType(
            #     {
            #         0: AxisType(BatchTag),
            #         1: AxisType(ChannelTag, 64),
            #         2: AxisType(HeightTag, 4),
            #         3: AxisType(WidthTag, 4),
            #     }
            # ),
            # "image": NeuralType(
            #     {
            #         0: AxisType(BatchTag),
            #         1: AxisType(ChannelTag),
            #         2: AxisType(HeightTag, self._input_size[1]),
            #         3: AxisType(WidthTag, self._input_size[0]),
            #     }
            # ),
            # "label": NeuralType({0: AxisType(BatchTag)}),
            "latent": NeuralType(('B', 'C', 'H', 'W'), ChannelType()),
            "image": NeuralType(('B', 'C', 'H', 'W'), ChannelType()),
            "label": NeuralType(tuple('B'), LabelsType()),
        }

    def __init__(self, batch_size, root, train=True, shuffle=True):
        super().__init__()
        self._input_size = (28, 28)

        self._batch_size = batch_size
        self._train = train
        self._shuffle = shuffle
        self._root = root
        self._transforms = transforms.Compose([transforms.ToTensor()])

        self._dataset = datasets.MNIST(root=self._root, train=self._train, download=True, transform=self._transforms)

        class DatasetWrapper(Dataset):
            def __init__(self, dataset):
                super().__init__()
                self._dataset = dataset

            def __getitem__(self, index):
                latents = torch.randn(64, 4, 4)
                items = self._dataset.__getitem__(index)
                return latents, items[0], items[1]

            def __len__(self):
                return self._dataset.__len__()

        self._dataset = DatasetWrapper(self._dataset)

    def __len__(self):
        return len(self._dataset)

    @property
    def dataset(self):
        return self._dataset

    @property
    def data_iterator(self):
        return None



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2020 The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import os

import torch
import wget
from transformers import TRANSFORMERS_CACHE, cached_path

__all__ = [
    'MEGATRON_CACHE',
    'MEGATRON_CONFIG_MAP',
    'CONFIGS',
    'get_megatron_lm_models_list',
    'get_megatron_config_file',
    'get_megatron_vocab_file',
    'get_megatron_checkpoint',
]

MEGATRON_CACHE = os.path.join(os.path.dirname(str(TRANSFORMERS_CACHE)), 'megatron')

CONFIGS = {'345m': {"hidden-size": 1024, "num-attention-heads": 16, "num-layers": 24, "max-seq-length": 512}}

MEGATRON_CONFIG_MAP = {
    'megatron-bert-345m-uncased': {
        'config': CONFIGS['345m'],
        'checkpoint': 'https://api.ngc.nvidia.com/v2/models/nvidia/megatron_bert_345m/versions/v0.0/files/release/mp_rank_00/model_optim_rng.pt',
        'vocab': 'https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-uncased-vocab.txt',
        'do_lower_case': True,
    },
    'megatron-bert-345m-cased': {
        'config': CONFIGS['345m'],
        'checkpoint': 'https://api.ngc.nvidia.com/v2/models/nvidia/megatron_bert_345m/versions/v0.1_cased/files/release/mp_rank_00/model_optim_rng.pt',
        'vocab': 'https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-cased-vocab.txt',
        'do_lower_case': False,
    },
    'megatron-bert-uncased': {
        'config': None,
        'checkpoint': None,
        'vocab': 'https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-uncased-vocab.txt',
        'do_lower_case': True,
    },
    'megatron-bert-cased': {
        'config': None,
        'vocab': 'https://s3.amazonaws.com/models.huggingface.co/bert/bert-large-cased-vocab.txt',
        'do_lower_case': False,
    },
}


def get_megatron_lm_models_list():
    '''
    Return the list of support Megatron models
    '''
    return list(MEGATRON_CONFIG_MAP.keys())


def get_megatron_config_file(pretrained_model_name):
    '''
    Returns model config file
    Args:
        pretrained_model_name (str): pretrained model name
    Returns:
        config (dict): contains model configuration: number of hidden layers, number of attention heads, etc
    '''
    return MEGATRON_CONFIG_MAP[pretrained_model_name]['config']


def get_megatron_vocab_file(pretrained_model_name):
    '''
    Gets vocabulary file from cache or downloads it
    Args:
        pretrained_model_name (str): pretrained model name
    Returns:
        path (str): path to the vocab file 
    '''
    url = MEGATRON_CONFIG_MAP[pretrained_model_name]['vocab']
    path = cached_path(url, cache_dir=MEGATRON_CACHE)
    return path


def get_megatron_checkpoint(pretrained_model_name):
    '''
    Gets checkpoint file from cache or downloads it
    Args:
        pretrained_model_name (str): pretrained model name
    Returns:
        path (str): path to model checkpoint
    '''
    url = MEGATRON_CONFIG_MAP[pretrained_model_name]['checkpoint']
    path = os.path.join(MEGATRON_CACHE, pretrained_model_name)

    if not os.path.exists(path):
        master_device = not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0
        if not os.path.exists(path):
            if master_device:
                wget.download(url, path)
            # wait until the master process downloads the file and writes it to the cache dir
            if torch.distributed.is_initialized():
                torch.distributed.barrier()

    return path


def is_lower_cased_megatron(pretrained_model_name):
    '''
    Returns if the megatron is cased or uncased
    Args:
        pretrained_model_name (str): pretrained model name
    Returns:
        do_lower_cased (bool): whether the model uses lower cased data
    '''
    return MEGATRON_CONFIG_MAP[pretrained_model_name]['do_lower_case']



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.data import LanguageModelingDataset
from nemo.collections.nlp.nm.data_layers.text_datalayer import TextDataLayer
from nemo.core import ChannelType, LabelsType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['LanguageModelingDataLayer']


class LanguageModelingDataLayer(TextDataLayer):
    """
    Data layer for standard language modeling task.

    Args:
        dataset (str): path to text document with data
        tokenizer (TokenizerSpec): tokenizer
        max_seq_length (int): maximum allowed length of the text segments
        batch_size (int): batch size
        batch_step (int): how many tokens to skip between two successive
            segments of text when constructing batches
        dataset_type (Dataset):
                the underlying dataset. Default: LanguageModelingDataset
        shuffle (bool): whether to shuffle data or not. Default: False.
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        input_ids: indices of tokens which constitute batches of text segments
        input_mask: bool tensor with 0s in place of tokens to be masked
        labels: indices of tokens which should be predicted from each of the
            corresponding tokens in input_ids; for left-to-right language
            modeling equals to input_ids shifted by 1 to the right
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_mask": NeuralType(('B', 'T'), ChannelType()),
            "labels": NeuralType(('B', 'T'), LabelsType()),
        }

    def __init__(
        self,
        dataset,
        tokenizer,
        max_seq_length,
        batch_size,
        batch_step=128,
        dataset_type=LanguageModelingDataset,
        shuffle=False,
    ):
        dataset_params = {
            'dataset': dataset,
            'tokenizer': tokenizer,
            'max_seq_length': max_seq_length,
            'batch_step': batch_step,
        }
        super().__init__(dataset_type, dataset_params, batch_size, shuffle=shuffle)
